# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import string
from functools import cached_property
from typing import Sequence, TypeVar

import equinox as eqx
import jax
import numpy as np
from jaxtyping import PRNGKeyArray

import haliax.partitioning

from ..axis import (
    Axis,
    AxisSelection,
    axis_name,
    axis_spec_to_shape_dict,
    replace_axis,
    selects_axis,
    unsize_axes,
    without_axes,
)
from ..core import NamedArray, named
from ..jax_utils import named_call
from ..random import uniform
from ..util import ensure_tuple

T = TypeVar("T")


class _ConvBase(eqx.Module):
    """
    Base class for Conv and ConvTranspose. Mostly just contains shared code.
    """

    Spatial: tuple[str | Axis, ...] = eqx.field(static=True)
    In: Axis = eqx.field(static=True)
    Out: Axis = eqx.field(static=True)
    weight: NamedArray = eqx.field()
    bias: NamedArray | None = eqx.field()

    def _lhs_dim_spec(self, batch_index, inputs):
        # the dim spec are single letters, for things like NCHW
        lhs_dim_spec = ""
        for i, ax in enumerate(inputs.axes):
            if i == batch_index:
                lhs_dim_spec += "N"
            elif ax.name == self.In.name:
                lhs_dim_spec += "C"
            else:
                index_in_spatial = _index_of_name(self.Spatial, ax.name)
                if index_in_spatial >= 0:
                    lhs_dim_spec += self._spatial_dim_short_names[index_in_spatial]
                else:
                    # shouldn't happen
                    raise ValueError(f"Unexpected axis {ax.name}")
        return lhs_dim_spec

    @cached_property
    def _spatial_dim_short_names(self) -> str:
        banned_letters = "NCIO"
        spec = ""
        for x in self.Spatial:
            name = axis_name(x)
            assert len(name) > 0
            if name[0] not in banned_letters and name[0] not in spec:
                spec += name[0]
            else:
                for x in string.ascii_uppercase:
                    if x not in spec and x not in banned_letters:
                        spec += x
                        break
                else:
                    raise RuntimeError("Too many spatial dimensions")

        return spec

    @cached_property
    def _weight_dim_spec(self) -> str:
        # dim spec in format jax.lax.conv_general_dilated expects
        # it's typically OIWH or similar but in case anyone does something funny with the weight array:
        # "OI" + self._spatial_dim_short_names
        spec = ""
        for ax in self.weight.axes:
            # check name for grouped convolutions
            if ax.name == self.In.name:
                spec += "I"
            elif ax == self.Out:
                spec += "O"
            else:
                index_in_spatial = _index_of_name(self.Spatial, ax.name)
                if index_in_spatial >= 0:
                    spec += self._spatial_dim_short_names[index_in_spatial]
                else:
                    # shouldn't happen
                    raise ValueError(f"Unexpected axis {ax.name}")
        return spec


# Based on Equinox's Conv class
class Conv(_ConvBase):
    """General N-dimensional convolution."""

    kernel_size: tuple[int, ...] = eqx.field(static=True)
    stride: tuple[int, ...] = eqx.field(static=True)
    padding: tuple[tuple[int, int], ...] = eqx.field(static=True)
    dilation: tuple[int, ...] = eqx.field(static=True)
    groups: int = eqx.field(static=True)

    @staticmethod
    def init(
        Spatial: AxisSelection,
        In: Axis,
        Out: Axis,
        kernel_size: int | Sequence[int],
        *,
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        key: PRNGKeyArray,
    ):
        """
        Args:
            Spatial: names of spatial dimensions
            In: Axis of input channels
            Out: Axis of output channels
            kernel_size: The size of the convolutional kernel.
            stride: The stride of the convolution. Can be a single number or a tuple
            padding: The amount of padding to apply before and after each spatial
                     dimension.
            dilation: The dilation of the convolution.
            groups: The number of groups to split the input channels into. Each
                    group is convolved separately with its own kernel.
            use_bias: Whether to add a bias after the convolution.
            key: Random key
        """
        Spatial = axis_spec_to_shape_dict(Spatial)
        if len(Spatial) == 0:
            raise ValueError("Spatial must have at least one element")

        kernel_size = _expand_and_check_shape(len(Spatial), kernel_size, "kernel_size")
        stride = _expand_and_check_shape(len(Spatial), stride, "stride")
        dilation = _expand_and_check_shape(len(Spatial), dilation, "dilation")
        padding = _convert_padding_spec(Spatial, padding)

        kernel_spec = tuple(Axis(axis_name(n), s) for n, s in zip(Spatial, kernel_size))
        in_spec = In.resize(In.size // groups)

        weight_key, bias_key = jax.random.split(key, 2)

        limit = 1 / np.sqrt(np.prod(kernel_size) * in_spec.size)

        weight = uniform(weight_key, (Out, in_spec, *kernel_spec), minval=-limit, maxval=limit)
        if use_bias:
            bias = uniform(bias_key, (Out,), minval=-limit, maxval=limit)
        else:
            bias = None

        return Conv(Spatial, In, Out, weight, bias, kernel_size, stride, padding, dilation, groups)

    @named_call
    def __call__(self, inputs, *, key: PRNGKeyArray | None = None):
        """
        Args:
            inputs (NamedArray): Input array
            key (PRNGKeyArray: Not used, compat with other modules

        Returns:
            (NamedArray): Output array, with shape similar to inputs except:
                - `Spatial` dimensions are reduced via the usual convolution formula
                - `In` is replaced with `Out`

        Notes:
            That formula is:
                `out_size = (in_size + 2 * padding - dilation * (kernel_size - 1) - 1) // stride + 1`
        """
        del key

        # check input
        for ax in self.Spatial:
            if not selects_axis(inputs.axes, ax):
                raise ValueError(f"Missing spatial axis {ax} in inputs: {inputs.axes}")

        if not selects_axis(inputs.axes, self.In):
            raise ValueError(f"Missing input axis {self.In} in inputs: {inputs.axes}")

        if selects_axis(inputs.axes, self.Out):
            raise ValueError(f"Output axis {self.Out} already in inputs: {inputs.axes}")

        # this is a bit subtle, but we need to make sure that the input is in the right order
        # and has the right set of dimensions

        # Constraints:
        # * at most 1 batch dimension (we'll vmap as necessary)
        # * at most 1 channel dimension (which we enforce for this module)
        # Spatial dimensions are reduced via the usual convolution formula, so we have to drop to names

        # identify batch dims, which get special treatment
        # jax's conv_general_dilated only supports exactly one batch dimension (not 0), so we vmap over any others.
        # We could choose instead to flatten them, but then we'd definitely lose sharding.
        batch_dims = without_axes(inputs.axes, unsize_axes(self.weight.axes))
        flat_inputs, unflatten = haliax.core.flatten_all_axes_but(
            inputs, "__batch__", batch_dims, reorder_to_front=True
        )
        x = self._do_conv(flat_inputs)
        x = unflatten(x)

        if self.bias is not None:
            x = x + self.bias

        return x

    def _do_conv(self, inputs):
        # _do_conv expects there to be a single __batch__ dimension
        output_axes = _compute_output_axes(inputs, "__batch__", self.In, self.Out)

        batch_index = _index_of_name(inputs.axes, "__batch__")

        lhs_dim_spec = self._lhs_dim_spec(batch_index, inputs)
        rhs_dim_spec = self._weight_dim_spec
        output_dim_spec = lhs_dim_spec
        x = jax.lax.conv_general_dilated(
            lhs=inputs.array,
            rhs=self.weight.array,
            window_strides=self.stride,
            padding=self.padding,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
            dimension_numbers=(lhs_dim_spec, rhs_dim_spec, output_dim_spec),
        )

        return named(x, output_axes)


class ConvTranspose(_ConvBase):
    """
    General N-dimensional transposed convolution.

    Based on Equinox's ConvTranspose class
    """

    kernel_size: tuple[int, ...] = eqx.field(static=True)
    stride: tuple[int, ...] = eqx.field(static=True)
    padding: tuple[tuple[int, int], ...] = eqx.field(static=True)
    output_padding: tuple[int, ...] = eqx.field(static=True)
    dilation: tuple[int, ...] = eqx.field(static=True)
    groups: int = eqx.field(static=True)

    @staticmethod
    def init(
        Spatial: AxisSelection,
        In: Axis,
        Out: Axis,
        kernel_size: int | Sequence[int],
        *,
        stride: int | Sequence[int] = 1,
        padding: int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        output_padding: int | Sequence[int] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        key: PRNGKeyArray,
    ):
        """

        Args:
            Spatial: Spatial dimensions
            In:  Input channels
            Out:  Output channels
            kernel_size: Kernel size, can be a single number or a tuple
            stride: Stride, can be a single number or a tuple
            padding: Padding, can be a single number or a tuple
            output_padding: Output padding, can be a single number or a tuple
            dilation: Dilation, can be a single number or a tuple
            groups: Number of groups to split the input channels into
            use_bias:  Whether to add on a bias after the convolution
            key:  Random key

        Notes:
            Output padding is the amount of extra padding to add to the output.
            Because the output size is not uniquely determined by the input size for transposed convolutions.
        """
        k_w, k_b = jax.random.split(key, 2)

        Spatial = axis_spec_to_shape_dict(Spatial)

        kernel_size = _expand_and_check_shape(len(Spatial), kernel_size, "kernel_size")
        stride = _expand_and_check_shape(len(Spatial), stride, "stride")
        dilation = _expand_and_check_shape(len(Spatial), dilation, "dilation")
        padding = _convert_padding_spec(Spatial, padding)
        output_padding = _expand_and_check_shape(len(Spatial), output_padding, "output_padding")

        kernel_spec = tuple(Axis(axis_name(n), s) for n, s in zip(Spatial, kernel_size))
        in_spec = In.resize(In.size // groups)

        lim = 1 / np.sqrt(np.prod(kernel_size) * in_spec.size)
        weight = uniform(k_w, (Out, in_spec, *kernel_spec), minval=-lim, maxval=lim)
        if use_bias:
            bias = uniform(k_b, (Out,), minval=-lim, maxval=lim)
        else:
            bias = None

        return ConvTranspose(
            Spatial,
            In,
            Out,
            weight,
            bias,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
        )

    @named_call
    def __call__(self, inputs, *, key: PRNGKeyArray | None = None):
        """
        Args:
            inputs (NamedArray): Input array
            key (PRNGKeyArray: Not used, compat with other modules

        Returns:
            (NamedArray): Output array, with shape similar to inputs except:
                - `Spatial` dimensions are increased via the usual (de)convolution formula
                - `In` is replaced with `Out`

        Notes:
            That formula is:
                `out_size = (in_size - 1) * stride - 2 * padding + dilation * (kernel_size - 1) + output_padding + 1`
        """
        del key

        batch_dims = without_axes(inputs.axes, unsize_axes(self.weight.axes))
        flat_inputs, unflatten = haliax.core.flatten_all_axes_but(
            inputs, "__batch__", batch_dims, reorder_to_front=True
        )
        x = self._do_conv(flat_inputs)
        x = unflatten(x)

        if self.bias is not None:
            x = x + self.bias

        output_axes = _compute_output_axes(inputs, batch_dims, self.In, self.Out)

        return x.rearrange(output_axes)

    def _do_conv(self, inputs):
        # cribbed from Equinox's ConvTranspose class
        padding = tuple(
            (d * (k - 1) - p0, d * (k - 1) - p1 + o)
            for k, (p0, p1), o, d in zip(self.kernel_size, self.padding, self.output_padding, self.dilation)
        )

        output_axes = _compute_output_axes(inputs, "__batch__", self.In, self.Out)
        batch_index = _index_of_name(inputs.axes, "__batch__")

        lhs_dim_spec = self._lhs_dim_spec(batch_index, inputs)
        rhs_dim_spec = self._weight_dim_spec
        output_dim_spec = lhs_dim_spec

        x = jax.lax.conv_general_dilated(
            lhs=inputs.array,
            rhs=self.weight.array,
            window_strides=(1,) * len(self.Spatial),
            padding=padding,
            lhs_dilation=self.stride,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
            dimension_numbers=(lhs_dim_spec, rhs_dim_spec, output_dim_spec),
        )

        return named(x, output_axes)


def _expand_and_check_shape(expected_len: int, spec: T | Sequence[T], name: str) -> tuple[T, ...]:
    spec = ensure_tuple(spec)
    if len(spec) == 1:
        spec = spec * expected_len
    if len(spec) != expected_len:
        raise ValueError(f"Expected {expected_len} elements for {name}, got {len(spec)}")

    return spec


def _convert_padding_spec(Spatial, padding):
    if isinstance(padding, int):
        padding = ((padding, padding),) * len(Spatial)
    elif isinstance(padding, tuple):
        padding = _expand_and_check_shape(len(Spatial), padding, "padding")
        padding_spec: list = []
        for p in padding:
            if isinstance(p, int):
                padding_spec.append((p, p))
            elif isinstance(p, tuple):
                padding_spec.append(p)
            else:
                raise ValueError(f"Invalid padding spec: {padding}")

        padding = tuple(padding_spec)
    else:
        raise ValueError(f"Invalid padding spec: {padding}")

    return padding


def _index_of_name(names: Sequence[str | Axis], name) -> int:
    for i, x in enumerate(names):
        if isinstance(x, Axis):
            x = axis_name(x)
        if x == name:
            return i
    return -1


def _compute_output_axes(inputs, batch_dims, In, Out):
    """
    Does two things:
    1. Replace In with Out
    2. turn spatial dims (non-batch, non-In, non-Out) into raw names b/c they change size in convolutions
    """
    unchanging_dims = [Out, *batch_dims]
    return [ax.name if ax not in unchanging_dims else ax for ax in replace_axis(inputs.axes, In, Out)]



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0

import jax.numpy as jnp
from typing import cast, no_type_check
import haliax as hax
from haliax import Axis
from haliax.core import NamedArray


def _assert_coefficients(coeffs: NamedArray, expected: jnp.ndarray, axis: Axis) -> None:
    assert jnp.allclose(coeffs.array, expected)  # type: ignore[union-attr]
    assert coeffs.axes[0] == axis.resize(coeffs.array.shape[0])  # type: ignore[union-attr]


def _assert_covariance_matrix(cov: NamedArray, axis: Axis, expected: jnp.ndarray) -> None:
    assert jnp.allclose(cov.array, expected)  # type: ignore[union-attr]
    assert cov.axes[0] == axis  # type: ignore[union-attr]
    assert cov.axes[1].name == f"{axis.name}_cov"  # type: ignore[union-attr]
    assert cov.axes[1].size == axis.size  # type: ignore[union-attr]


def test_poly():
    R = Axis("R", 3)
    roots = hax.named([1.0, 2.0, 3.0], (R,))
    coeffs = hax.poly(roots)
    assert jnp.allclose(coeffs.array, jnp.poly(roots.array))
    assert coeffs.axes[0] == R.resize(coeffs.array.shape[0])


def test_poly_arithmetic():
    C1 = Axis("C1", 4)
    p = hax.named([1.0, 0.0, -2.0, 1.0], (C1,))
    C2 = Axis("C2", 2)
    q = hax.named([1.0, -1.0], (C2,))

    add = hax.polyadd(p, q)
    assert jnp.allclose(add.array, jnp.polyadd(p.array, q.array))
    assert add.axes[0] == C1.resize(add.array.shape[0])

    sub = hax.polysub(p, q)
    assert jnp.allclose(sub.array, jnp.polysub(p.array, q.array))
    assert sub.axes[0] == C1.resize(sub.array.shape[0])

    mul = hax.polymul(p, q)
    assert jnp.allclose(mul.array, jnp.polymul(p.array, q.array))
    assert mul.axes[0] == C1.resize(mul.array.shape[0])

    div_q, div_r = hax.polydiv(p, q)
    exp_q, exp_r = jnp.polydiv(p.array, q.array)
    assert jnp.allclose(div_q.array, exp_q)
    assert jnp.allclose(div_r.array, exp_r)
    assert div_q.axes[0] == C1.resize(div_q.array.shape[0])
    assert div_r.axes[0] == C1.resize(div_r.array.shape[0])


def test_poly_int_der():
    C = Axis("C", 4)
    p = hax.named([1.0, 0.0, -2.0, 1.0], (C,))

    d = hax.polyder(p)
    i = hax.polyint(p)

    assert jnp.allclose(d.array, jnp.polyder(p.array))
    assert jnp.allclose(i.array, jnp.polyint(p.array))
    assert d.axes[0] == C.resize(d.array.shape[0])
    assert i.axes[0] == C.resize(i.array.shape[0])


@no_type_check
def test_polyval_polyfit():
    C = Axis("C", 4)
    p = hax.named([1.0, 0.0, -2.0, 1.0], (C,))

    X = Axis("X", 5)
    x = hax.arange(X, dtype=jnp.float64)
    y = hax.polyval(p, x)

    pv = cast(NamedArray, hax.polyval(p, x))
    assert jnp.allclose(pv.array, jnp.polyval(p.array, x.array))

    fit = cast(NamedArray, hax.polyfit(x, y, 3))
    assert jnp.allclose(fit.array, p.array, atol=1e-5)
    assert fit.axes[0] == X.resize(fit.array.shape[0])

    full_fit = hax.polyfit(x, y, 3, full=True)
    assert isinstance(full_fit, tuple)
    coeffs_full = full_fit[0]
    assert isinstance(coeffs_full, NamedArray)
    coeffs_full = cast(NamedArray, coeffs_full)
    exp_full = jnp.polyfit(x.array, y.array, 3, full=True)
    _assert_coefficients(coeffs_full, exp_full[0], X)  # type: ignore[union-attr]
    for res, exp in zip(full_fit[1:], exp_full[1:]):
        assert jnp.allclose(res, exp)

    cov_fit = hax.polyfit(x, y, 3, cov=True)
    assert isinstance(cov_fit, tuple)
    coeffs_cov = cov_fit[0]
    cov = cov_fit[1]
    assert isinstance(coeffs_cov, NamedArray)
    assert isinstance(cov, NamedArray)
    coeffs_cov = cast(NamedArray, coeffs_cov)
    cov = cast(NamedArray, cov)
    exp_coeffs, exp_cov = jnp.polyfit(x.array, y.array, 3, cov=True)
    _assert_coefficients(coeffs_cov, exp_coeffs, X)  # type: ignore[union-attr]
    _assert_covariance_matrix(cov, coeffs_cov.axes[0], exp_cov)  # type: ignore[union-attr]


def test_poly_roots_trim_vander():
    C = Axis("C", 4)
    p = hax.named([1.0, 0.0, -2.0, 1.0], (C,))

    r = hax.roots(p)
    assert jnp.allclose(r.array, jnp.roots(p.array))
    assert r.axes[0] == C.resize(r.array.shape[0])

    C2 = Axis("C2", 4)
    p2 = hax.named([0.0, 0.0, 1.0, 2.0], (C2,))
    t = hax.trim_zeros(p2)
    assert jnp.allclose(t.array, jnp.trim_zeros(p2.array))
    assert t.axes[0] == C2.resize(t.array.shape[0])

    X = Axis("X", 3)
    x = hax.named([1.0, 2.0, 3.0], (X,))
    D = Axis("D", 3)
    v = hax.vander(x, D)
    assert jnp.allclose(v.array, jnp.vander(x.array, N=3))
    assert v.axes == (X, D)

    v2 = hax.vander(x, "E")
    assert jnp.allclose(v2.array, jnp.vander(x.array))
    assert v2.axes[0] == X
    assert v2.axes[1] == Axis("E", X.size)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import jax.numpy as jnp
import equinox as eqx
import pytest

import haliax as hax


class M(eqx.Module):
    a: jnp.ndarray = hax.field(axis_names=("batch",))


def test_axis_names_metadata():
    field = M.__dataclass_fields__["a"]
    assert field.metadata["axis_names"] == ("batch",)


def test_axis_names_static_exclusive():
    with pytest.raises(ValueError):
        hax.field(static=True, axis_names=("x",))



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import typing
from typing import Mapping

import jax
import jax.numpy as jnp
from jaxtyping import ArrayLike

import haliax

from .axis import Axis, AxisSelector, axis_name
from .core import NamedArray, NamedOrNumeric, broadcast_arrays, broadcast_arrays_and_return_axes, named
from .jax_utils import ensure_scalar, is_scalarish


def trace(array: NamedArray, axis1: AxisSelector, axis2: AxisSelector, offset=0, dtype=None) -> NamedArray:
    """Compute the trace of an array along two named axes."""
    a1_index = array.axis_indices(axis1)
    a2_index = array.axis_indices(axis2)

    if a1_index is None:
        raise ValueError(f"Axis {axis1} not found in array. Available axes: {array.axes}")
    if a2_index is None:
        raise ValueError(f"Axis {axis2} not found in array. Available axes: {array.axes}")

    if a1_index == a2_index:
        raise ValueError(f"Cannot trace along the same axis. Got {axis1} and {axis2}")

    inner = jnp.trace(array.array, offset=offset, axis1=a1_index, axis2=a2_index, dtype=dtype)
    # remove the two indices
    axes = tuple(a for i, a in enumerate(array.axes) if i not in (a1_index, a2_index))
    return NamedArray(inner, axes)


@typing.overload
def where(
    condition: NamedOrNumeric | bool,
    x: NamedOrNumeric,
    y: NamedOrNumeric,
) -> NamedArray: ...


@typing.overload
def where(
    condition: NamedArray,
    *,
    fill_value: int,
    new_axis: Axis,
) -> tuple[NamedArray, ...]: ...


def where(
    condition: NamedOrNumeric | bool,
    x: NamedOrNumeric | None = None,
    y: NamedOrNumeric | None = None,
    fill_value: int | None = None,
    new_axis: Axis | None = None,
) -> NamedArray | tuple[NamedArray, ...]:
    """Like jnp.where, but with named axes."""

    if (x is None) != (y is None):
        raise ValueError("Must either specify both x and y, or neither")

    # one argument form
    if (x is None) and (y is None):
        if not isinstance(condition, NamedArray):
            raise ValueError(f"condition {condition} must be a NamedArray in single argument mode")
        if fill_value is None or new_axis is None:
            raise ValueError("Must specify both fill_value and new_axis")
        return tuple(
            NamedArray(idx, (new_axis,))
            for idx in jnp.where(condition.array, size=new_axis.size, fill_value=fill_value)
        )

    # if x or y is a NamedArray, the other must be as well. wrap as needed for scalars

    if is_scalarish(condition):
        if x is None or y is None:
            raise ValueError("Must specify x and y when condition is a scalar")

        if isinstance(x, NamedArray) and not isinstance(y, NamedArray):
            if not is_scalarish(y):
                raise ValueError("y must be a NamedArray or scalar if x is a NamedArray")
            y = named(y, ())
        elif isinstance(y, NamedArray) and not isinstance(x, NamedArray):
            if not is_scalarish(x):
                raise ValueError("x must be a NamedArray or scalar if y is a NamedArray")
            x = named(x, ())
        x, y = broadcast_arrays(x, y)
        if isinstance(condition, NamedArray):
            condition = ensure_scalar(condition, name="condition")
        return jax.lax.cond(condition, lambda _: x, lambda _: y, None)

    condition, x, y = broadcast_arrays(condition, x, y)  # type: ignore

    assert isinstance(condition, NamedArray)

    def _array_if_named(x):
        if isinstance(x, NamedArray):
            return x.array
        return x

    raw = jnp.where(condition.array, _array_if_named(x), _array_if_named(y))
    return NamedArray(raw, condition.axes)


def nonzero(array: NamedArray, *, size: Axis, fill_value: int = 0) -> tuple[NamedArray, ...]:
    """Like :func:`jax.numpy.nonzero`, but with named axes.

    Args:
        array: The input array to test for nonzero values. Must be a :class:`NamedArray`.
        size: Axis specifying the size of the output axis. This is required because
            JAX requires the size of the result at tracing time.
        fill_value: Value used to fill the output when fewer than ``size`` elements are
            nonzero. Defaults to ``0``.

    Returns:
        A tuple of :class:`NamedArray` objects, one for each axis of ``array``. Each
        returned array has ``size`` as its only axis and contains the indices of the
        nonzero elements along the corresponding input axis.
    """

    if not isinstance(array, NamedArray):
        raise ValueError("array must be a NamedArray")

    return tuple(NamedArray(idx, (size,)) for idx in jnp.nonzero(array.array, size=size.size, fill_value=fill_value))


def clip(array: NamedOrNumeric, a_min: NamedOrNumeric, a_max: NamedOrNumeric) -> NamedArray:
    """Like jnp.clip, but with named axes. This version currently only accepts the three argument form."""
    (array, a_min, a_max), axes = broadcast_arrays_and_return_axes(array, a_min, a_max)
    array = raw_array_or_scalar(array)
    a_min = raw_array_or_scalar(a_min)
    a_max = raw_array_or_scalar(a_max)

    return NamedArray(jnp.clip(array, a_min, a_max), axes)


def tril(array: NamedArray, axis1: Axis, axis2: Axis, k=0) -> NamedArray:
    """Compute the lower triangular part of an array along two named axes."""
    array = array.rearrange((..., axis1, axis2))

    inner = jnp.tril(array.array, k=k)
    return NamedArray(inner, array.axes)


def triu(array: NamedArray, axis1: Axis, axis2: Axis, k=0) -> NamedArray:
    """Compute the upper triangular part of an array along two named axes."""
    array = array.rearrange((..., axis1, axis2))

    inner = jnp.triu(array.array, k=k)
    return NamedArray(inner, array.axes)


def isclose(a: NamedArray, b: NamedArray, rtol=1e-05, atol=1e-08, equal_nan=False) -> NamedArray:
    """Returns a boolean array where two arrays are element-wise equal within a tolerance."""
    a, b = broadcast_arrays(a, b)
    # TODO: numpy supports an array atol and rtol, but we don't yet
    return NamedArray(jnp.isclose(a.array, b.array, rtol=rtol, atol=atol, equal_nan=equal_nan), a.axes)


def allclose(a: NamedArray, b: NamedArray, rtol=1e-05, atol=1e-08, equal_nan=False) -> bool:
    """Returns True if two arrays are element-wise equal within a tolerance."""
    a, b = broadcast_arrays(a, b)
    return bool(jnp.allclose(a.array, b.array, rtol=rtol, atol=atol, equal_nan=equal_nan))


def array_equal(a: NamedArray, b: NamedArray) -> bool:
    """Returns True if two arrays have the same shape and elements."""
    if set(a.axes) != set(b.axes):
        return False
    b = b.rearrange(a.axes)
    return bool(jnp.array_equal(a.array, b.array))


def array_equiv(a: NamedArray, b: NamedArray) -> bool:
    """Returns True if two arrays are shape-consistent and equal."""
    try:
        a, b = broadcast_arrays(a, b)
    except ValueError:
        return False
    return bool(jnp.array_equal(a.array, b.array))


def pad_left(array: NamedArray, axis: Axis, new_axis: Axis, value=0) -> NamedArray:
    """Pad an array along named axes."""
    amount_to_pad_to = new_axis.size - axis.size
    if amount_to_pad_to < 0:
        raise ValueError(f"Cannot pad {axis} to {new_axis}")

    idx = array.axis_indices(axis)

    padding = [(0, 0)] * array.ndim
    if idx is None:
        raise ValueError(f"Axis {axis} not found in array. Available axes: {array.axes}")
    padding[idx] = (amount_to_pad_to, 0)

    padded = jnp.pad(array.array, padding, constant_values=value)
    return NamedArray(padded, array.axes[:idx] + (new_axis,) + array.axes[idx + 1 :])


def pad(
    array: NamedArray,
    pad_width: Mapping[AxisSelector, tuple[int, int]],
    *,
    mode: str = "constant",
    constant_values: NamedOrNumeric = 0,
    **kwargs,
) -> NamedArray:
    """Version of ``jax.numpy.pad`` that works with ``NamedArray``.

    ``pad_width`` should be a mapping from axis (or axis name) to a ``(before, after)``
    tuple specifying how much padding to add on each side of that axis. Any axis
    not present in ``pad_width`` will not be padded.
    """

    padding = []
    new_axes = []
    for ax in array.axes:
        left_right = pad_width.get(ax)
        if left_right is None:
            left_right = pad_width.get(axis_name(ax))  # type: ignore[arg-type]
        if left_right is None:
            left_right = (0, 0)
        left, right = left_right
        padding.append((left, right))
        new_axes.append(ax.resize(ax.size + left + right))

    result = jnp.pad(
        array.array,
        padding,
        mode=mode,
        constant_values=raw_array_or_scalar(constant_values),
        **kwargs,
    )

    return NamedArray(result, tuple(new_axes))


def raw_array_or_scalar(x: NamedOrNumeric):
    if isinstance(x, NamedArray):
        return x.array
    return x


@typing.overload
def unique(
    array: NamedArray, Unique: Axis, *, axis: AxisSelector | None = None, fill_value: ArrayLike | None = None
) -> NamedArray: ...


@typing.overload
def unique(
    array: NamedArray,
    Unique: Axis,
    *,
    return_index: typing.Literal[True],
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray]: ...


@typing.overload
def unique(
    array: NamedArray,
    Unique: Axis,
    *,
    return_inverse: typing.Literal[True],
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray]: ...


@typing.overload
def unique(
    array: NamedArray,
    Unique: Axis,
    *,
    return_counts: typing.Literal[True],
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray]: ...


@typing.overload
def unique(
    array: NamedArray,
    Unique: Axis,
    *,
    return_index: bool = False,
    return_inverse: bool = False,
    return_counts: bool = False,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> NamedArray | tuple[NamedArray, ...]: ...


def unique(
    array: NamedArray,
    Unique: Axis,
    *,
    return_index: bool = False,
    return_inverse: bool = False,
    return_counts: bool = False,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> NamedArray | tuple[NamedArray, ...]:
    """
    Like jnp.unique, but with named axes.

    Args:
        array: The input array.
        Unique: The name of the axis that will be created to hold the unique values.
        fill_value: The value to use for the fill_value argument of jnp.unique
        axis: The axis along which to find unique values.
        return_index: If True, return the indices of the unique values.
        return_inverse: If True, return the indices of the input array that would reconstruct the unique values.
    """
    size = Unique.size

    is_multireturn = return_index or return_inverse or return_counts

    kwargs = dict(
        size=size,
        fill_value=fill_value,
        return_index=return_index,
        return_inverse=return_inverse,
        return_counts=return_counts,
    )

    if axis is not None:
        axis_index = array.axis_indices(axis)
        if axis_index is None:
            raise ValueError(f"Axis {axis} not found in array. Available axes: {array.axes}")
        out = jnp.unique(array.array, axis=axis_index, **kwargs)
    else:
        out = jnp.unique(array.array, **kwargs)

    if is_multireturn:
        unique = out[0]
        next_index = 1
        if return_index:
            index = out[next_index]
            next_index += 1
        if return_inverse:
            inverse = out[next_index]
            next_index += 1
        if return_counts:
            counts = out[next_index]
            next_index += 1
    else:
        unique = out

    ret = []

    if axis is not None:
        out_axes = haliax.axis.replace_axis(array.axes, axis, Unique)
    else:
        out_axes = (Unique,)

    unique_values = haliax.named(unique, out_axes)
    if not is_multireturn:
        return unique_values

    ret.append(unique_values)

    if return_index:
        ret.append(haliax.named(index, Unique))

    if return_inverse:
        if axis is not None:
            assert axis_index is not None
            inverse = haliax.named(inverse, array.axes[axis_index])
        else:
            inverse = haliax.named(inverse, array.axes)
        ret.append(inverse)

    if return_counts:
        ret.append(haliax.named(counts, Unique))

    return tuple(ret)


def unique_values(
    array: NamedArray,
    Unique: Axis,
    *,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> NamedArray:
    """Shortcut for :func:`unique` that returns only unique values."""

    return typing.cast(
        NamedArray,
        unique(
            array,
            Unique,
            axis=axis,
            fill_value=fill_value,
        ),
    )


def unique_counts(
    array: NamedArray,
    Unique: Axis,
    *,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray]:
    """Shortcut for :func:`unique` that also returns counts."""

    values, counts = typing.cast(
        tuple[NamedArray, NamedArray],
        unique(
            array,
            Unique,
            return_counts=True,
            axis=axis,
            fill_value=fill_value,
        ),
    )
    return values, counts


def unique_inverse(
    array: NamedArray,
    Unique: Axis,
    *,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray]:
    """Shortcut for :func:`unique` that also returns inverse indices."""

    values, inverse = typing.cast(
        tuple[NamedArray, NamedArray],
        unique(
            array,
            Unique,
            return_inverse=True,
            axis=axis,
            fill_value=fill_value,
        ),
    )
    return values, inverse


def unique_all(
    array: NamedArray,
    Unique: Axis,
    *,
    axis: AxisSelector | None = None,
    fill_value: ArrayLike | None = None,
) -> tuple[NamedArray, NamedArray, NamedArray, NamedArray]:
    """Shortcut for :func:`unique` returning values, indices, inverse, and counts."""

    values, indices, inverse, counts = typing.cast(
        tuple[NamedArray, NamedArray, NamedArray, NamedArray],
        unique(
            array,
            Unique,
            return_index=True,
            return_inverse=True,
            return_counts=True,
            axis=axis,
            fill_value=fill_value,
        ),
    )
    return values, indices, inverse, counts


def searchsorted(
    a: NamedArray,
    v: NamedArray | ArrayLike,
    *,
    side: str = "left",
    sorter: NamedArray | ArrayLike | None = None,
    method: str = "scan",
) -> NamedArray:
    """Named version of `jax.numpy.searchsorted`.

    ``a`` and ``sorter`` (if provided) must be one-dimensional.
    The returned array has the same axes as ``v``.
    """

    if a.ndim != 1:
        raise ValueError("searchsorted only supports 1D 'a'")

    if not isinstance(v, NamedArray):
        v = haliax.named(v, ())

    sorter_arr = None
    if sorter is not None:
        sorter_arr = sorter.array if isinstance(sorter, NamedArray) else jnp.asarray(sorter)

    result = jnp.searchsorted(a.array, v.array, side=side, sorter=sorter_arr, method=method)
    return NamedArray(result, v.axes)


def bincount(
    x: NamedArray,
    Counts: Axis,
    *,
    weights: NamedArray | ArrayLike | None = None,
    minlength: int = 0,
) -> NamedArray:
    """Named version of `jax.numpy.bincount`.

    The output axis is specified by ``Counts``.
    """

    if x.ndim != 1:
        raise ValueError("bincount only supports 1D arrays")

    w_array = None
    if weights is not None:
        if isinstance(weights, NamedArray):
            weights = haliax.broadcast_to(weights, x.axes)
            w_array = weights.array
        else:
            w_array = jnp.asarray(weights)

    result = jnp.bincount(x.array, weights=w_array, minlength=minlength, length=Counts.size)
    return NamedArray(result, (Counts,))


def packbits(a: NamedArray, axis: AxisSelector, *, bitorder: str = "big") -> NamedArray:
    """Named version of `jax.numpy.packbits`."""

    axis_index = a.axis_indices(axis)
    if not isinstance(axis_index, int):
        raise ValueError("packbits only supports a single existing axis")

    result = jnp.packbits(a.array, axis=axis_index, bitorder=bitorder)
    old_axis = a.axes[axis_index]
    new_size = (old_axis.size + 7) // 8
    new_axis = old_axis.resize(new_size)
    new_axes = a.axes[:axis_index] + (new_axis,) + a.axes[axis_index + 1 :]
    return NamedArray(result, new_axes)


def unpackbits(
    a: NamedArray,
    axis: AxisSelector,
    *,
    count: int | None = None,
    bitorder: str = "big",
) -> NamedArray:
    """Named version of `jax.numpy.unpackbits`."""

    axis_index = a.axis_indices(axis)
    if not isinstance(axis_index, int):
        raise ValueError("unpackbits only supports a single existing axis")

    result = jnp.unpackbits(a.array, axis=axis_index, count=count, bitorder=bitorder)
    old_axis = a.axes[axis_index]
    new_size = count if count is not None else old_axis.size * 8
    new_axis = old_axis.resize(new_size)
    new_axes = a.axes[:axis_index] + (new_axis,) + a.axes[axis_index + 1 :]
    return NamedArray(result, new_axes)


__all__ = [
    "trace",
    "where",
    "tril",
    "triu",
    "isclose",
    "pad_left",
    "pad",
    "clip",
    "packbits",
    "unpackbits",
    "unique",
    "unique_values",
    "unique_counts",
    "unique_inverse",
    "unique_all",
    "searchsorted",
    "bincount",
]



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from typing import Callable
import typing

import jax.numpy as jnp
import pytest
from jax.random import PRNGKey

import haliax as hax
from haliax import Axis, NamedArray


def test_trace():
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)
    named1 = hax.random.uniform(PRNGKey(0), (Width, Depth))
    trace1 = hax.trace(named1, Width, Depth)
    assert jnp.all(jnp.isclose(trace1.array, jnp.trace(named1.array)))
    assert len(trace1.axes) == 0

    trace1 = hax.trace(named1, "Width", "Depth")
    assert jnp.all(jnp.isclose(trace1.array, jnp.trace(named1.array)))
    assert len(trace1.axes) == 0

    Height = Axis("Height", 10)
    named2 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    trace2 = hax.trace(named2, Width, Depth)
    assert jnp.all(jnp.isclose(trace2.array, jnp.trace(named2.array, axis1=1, axis2=2)))
    assert trace2.axes == (Height,)

    trace2 = hax.trace(named2, "Width", "Depth")
    assert jnp.all(jnp.isclose(trace2.array, jnp.trace(named2.array, axis1=1, axis2=2)))
    assert trace2.axes == (Height,)


def test_add():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named2 = hax.random.uniform(PRNGKey(1), (Height, Width, Depth))

    named3 = named1 + named2
    assert jnp.all(jnp.isclose(named3.array, named1.array + named2.array))

    named2_reorder = named2.rearrange((Width, Height, Depth))
    named4 = named1 + named2_reorder
    named4 = named4.rearrange((Height, Width, Depth))
    assert jnp.all(jnp.isclose(named4.array, named1.array + named2.array))


def test_add_broadcasting():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named2 = hax.random.uniform(PRNGKey(1), (Width, Depth))

    named3 = named1 + named2
    assert jnp.all(jnp.isclose(named3.array, named1.array + named2.array))

    named2_reorder = named2.rearrange((Depth, Width))
    named4 = named1 + named2_reorder
    named4 = named4.rearrange((Height, Width, Depth))

    assert jnp.all(jnp.isclose(named4.array, named1.array + named2.array))

    # now for the broadcasting we don't like
    named5 = hax.random.uniform(PRNGKey(1), (Height, Depth))
    named6 = hax.random.uniform(PRNGKey(2), (Width, Depth))

    with pytest.raises(ValueError):
        _ = named5 + named6


def test_add_scalar():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named2 = named1 + 1.0
    assert jnp.all(jnp.isclose(named2.array, named1.array + 1.0))

    named3 = 1.0 + named1
    assert jnp.all(jnp.isclose(named3.array, named1.array + 1.0))


def test_add_no_overlap():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    named1: NamedArray = hax.random.uniform(PRNGKey(0), (Height))
    named2 = hax.random.uniform(PRNGKey(1), (Width, Depth))

    with pytest.raises(ValueError):
        _ = named1 + named2

    named3 = named1.broadcast_to((Height, Width, Depth)) + named2

    assert jnp.all(
        jnp.isclose(named3.array, named1.array.reshape((-1, 1, 1)) + named2.array.reshape((1,) + named2.array.shape))
    )


# TODO: tests for other ops:


@pytest.mark.parametrize("use_jit", [False, True])
def test_where(use_jit):
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    hax_where: Callable = hax.where
    if use_jit:
        hax_where = hax.named_jit(hax_where)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named2 = hax.random.uniform(PRNGKey(1), (Height, Width, Depth))

    hax_where(0.0, named1, 0.0)

    named3 = hax_where(named1 > named2, named1, named2)

    assert jnp.all(jnp.isclose(named3.array, jnp.where(named1.array > named2.array, named1.array, named2.array)))

    named2_reorder = named2.rearrange((Width, Height, Depth))
    named4 = hax_where(named1 > named2_reorder, named1, named2_reorder)
    named4 = named4.rearrange((Height, Width, Depth))
    assert jnp.all(jnp.isclose(named4.array, jnp.where(named1.array > named2.array, named1.array, named2.array)))

    # now some broadcasting
    named5 = hax.random.uniform(PRNGKey(1), (Height, Width))
    named6 = hax.random.uniform(PRNGKey(2), Width)

    named7 = hax_where(named5 > named6, named5, named6)
    named7 = named7.rearrange((Height, Width))
    assert jnp.all(jnp.isclose(named7.array, jnp.where(named5.array > named6.array, named5.array, named6.array)))

    # now for the broadcasting we don't like
    named5 = hax.random.uniform(PRNGKey(1), (Height, Depth))
    named6 = hax.random.uniform(PRNGKey(2), (Width, Depth))

    with pytest.raises(ValueError):
        _ = hax_where(named5 > named6, named5, named6)

    # now single argument mode
    Volume = hax.Axis("Volume", Height.size * Width.size * Depth.size)
    named7 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named8, named9, named10 = hax_where(named7 > 0.5, fill_value=-1, new_axis=Volume)
    unnamed_7 = named7.array
    unnamed_8, unnamed_9, unnamed_10 = jnp.where(unnamed_7 > 0.5, size=Volume.size, fill_value=-1)
    assert jnp.all(unnamed_8 == named8.array)


@pytest.mark.parametrize("use_jit", [False, True])
def test_nonzero(use_jit):
    Height = Axis("Height", 3)
    Width = Axis("Width", 4)

    data = jnp.array([[0, 1, 0, 2], [3, 0, 4, 0], [0, 0, 5, 6]])
    named = hax.named(data, (Height, Width))

    Nonzero = Axis("Nonzero", Height.size * Width.size)

    hax_nonzero: Callable = hax.nonzero
    if use_jit:
        hax_nonzero = hax.named_jit(hax_nonzero)

    i, j = hax_nonzero(named, size=Nonzero, fill_value=-1)
    ui, uj = jnp.nonzero(data, size=Nonzero.size, fill_value=-1)

    assert i.axes == (Nonzero,)
    assert j.axes == (Nonzero,)
    assert jnp.all(i.array == ui)
    assert jnp.all(j.array == uj)


def test_clip():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
    named2 = hax.clip(named1, 0.3, 0.7)
    assert jnp.all(jnp.isclose(named2.array, jnp.clip(named1.array, 0.3, 0.7)))

    named2_reorder = named2.rearrange((Width, Height, Depth))
    named3 = hax.clip(named2_reorder, 0.3, 0.7)
    named3 = named3.rearrange((Height, Width, Depth))
    assert jnp.all(jnp.isclose(named3.array, jnp.clip(named2.array, 0.3, 0.7)))

    # now some interesting broadcasting
    lower = hax.full((Height, Width), 0.3)
    upper = hax.full((Width, Depth), 0.7)
    named4 = hax.clip(named1, lower, upper)
    named4 = named4.rearrange((Height, Width, Depth))

    assert jnp.all(
        jnp.isclose(
            named4.array,
            jnp.clip(
                named1.array,
                lower.array.reshape((Height.size, Width.size, 1)),
                upper.array.reshape((1, Width.size, Depth.size)),
            ),
        )
    )


def test_tril_triu():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)
    Depth = Axis("Depth", 4)

    for hax_fn, jnp_fn in [(hax.tril, jnp.tril), (hax.triu, jnp.triu)]:
        named1 = hax.random.uniform(PRNGKey(0), (Height, Width, Depth))
        named2 = hax_fn(named1, Width, Depth)
        assert jnp.all(jnp.isclose(named2.array, jnp_fn(named1.array)))

        named3 = hax_fn(named1, Width, Depth, k=1)
        assert jnp.all(jnp.isclose(named3.array, jnp_fn(named1.array, k=1)))

        named4 = hax_fn(named1, Width, Depth, k=-1)
        assert jnp.all(jnp.isclose(named4.array, jnp_fn(named1.array, k=-1)))

        named5 = hax_fn(named1, Height, Depth)
        expected5 = jnp_fn(named1.array.transpose([1, 0, 2]))
        assert jnp.all(jnp.isclose(named5.array, expected5))


def test_mean_respects_where():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width))
    where = hax.random.uniform(PRNGKey(1), (Height, Width)) > 0.5

    assert not hax.all(hax.isclose(hax.mean(named1), hax.mean(named1, where=where)))
    assert jnp.all(jnp.isclose(hax.mean(named1, where=where).array, jnp.mean(named1.array, where=where.array)))

    # check broadcasting
    where = hax.random.uniform(PRNGKey(2), (Height,)) > 0.5
    assert not jnp.all(jnp.isclose(hax.mean(named1).array, hax.mean(named1, where=where).array))
    assert jnp.all(
        jnp.isclose(hax.mean(named1, where=where).array, jnp.mean(named1.array, where=where.array.reshape((-1, 1))))
    )


def test_reductions_produce_scalar_named_arrays_when_None_axis():
    Height = Axis("Height", 10)
    Width = Axis("Width", 3)

    named1 = hax.random.uniform(PRNGKey(0), (Height, Width))

    assert isinstance(hax.mean(named1, axis=None), NamedArray)

    # But if we specify axes, we always get a NamedArray, even if it's a scalar
    assert isinstance(hax.mean(named1, axis=("Height", "Width")), NamedArray)
    assert hax.mean(named1, axis=("Height", "Width")).axes == ()


def test_pad():
    Height = Axis("Height", 3)
    Width = Axis("Width", 2)

    arr = hax.arange((Height, Width))
    padded = hax.pad(arr, {Height: (1, 2), Width: (0, 1)}, mode="constant", constant_values=0)

    expected = jnp.pad(arr.array, [(1, 2), (0, 1)], mode="constant", constant_values=0)
    assert padded.axes[0].size == Height.size + 3
    assert padded.axes[1].size == Width.size + 1
    assert jnp.all(expected == padded.array)


def test_unique():
    # named version of this test:
    # >>> M = jnp.array([[1, 2],
    # ...                [2, 3],
    # ...                [1, 2]])
    # >>> jnp.unique(M)
    # Array([1, 2, 3], dtype=int32)

    Height = Axis("Height", 3)
    Width = Axis("Width", 2)

    named1 = hax.named([[1, 2], [2, 3], [1, 2]], (Height, Width))

    U = Axis("U", 3)

    named2 = hax.unique(named1, U)

    assert jnp.all(jnp.equal(named2.array, jnp.array([1, 2, 3])))

    #     If you pass an ``axis`` keyword, you can find unique *slices* of the array along
    #     that axis:
    #
    #     >>> jnp.unique(M, axis=0)
    #     Array([[1, 2],
    #            [2, 3]], dtype=int32)

    U2 = Axis("U2", 2)
    named3 = hax.unique(named1, U2, axis=Height)
    assert jnp.all(jnp.equal(named3.array, jnp.array([[1, 2], [2, 3]])))

    #     >>> x = jnp.array([3, 4, 1, 3, 1])
    #     >>> values, indices = jnp.unique(x, return_index=True)
    #     >>> print(values)
    #     [1 3 4]
    #     >>> print(indices)
    #     [2 0 1]
    #     >>> jnp.all(values == x[indices])
    #     Array(True, dtype=bool)

    x = hax.named([3, 4, 1, 3, 1], ("Height",))
    U3 = Axis("U3", 3)
    values, indices = hax.unique(x, U3, return_index=True)

    assert jnp.all(jnp.equal(values.array, jnp.array([1, 3, 4])))
    assert jnp.all(jnp.equal(indices.array, jnp.array([2, 0, 1])))

    assert jnp.all(jnp.equal(values.array, x[{"Height": indices}].array))

    #   If you set ``return_inverse=True``, then ``unique`` returns the indices within the
    #     unique values for every entry in the input array:
    #
    #     >>> x = jnp.array([3, 4, 1, 3, 1])
    #     >>> values, inverse = jnp.unique(x, return_inverse=True)
    #     >>> print(values)
    #     [1 3 4]
    #     >>> print(inverse)
    #     [1 2 0 1 0]
    #     >>> jnp.all(values[inverse] == x)
    #     Array(True, dtype=bool)

    values, inverse = hax.unique(x, U3, return_inverse=True)

    assert jnp.all(jnp.equal(values.array, jnp.array([1, 3, 4])))
    assert jnp.all(jnp.equal(inverse.array, jnp.array([1, 2, 0, 1, 0])))

    #     In multiple dimensions, the input can be reconstructed using
    #     :func:`jax.numpy.take`:
    #
    #     >>> values, inverse = jnp.unique(M, axis=0, return_inverse=True)
    #     >>> jnp.all(jnp.take(values, inverse, axis=0) == M)
    #     Array(True, dtype=bool)
    #

    values, inverse = hax.unique(named1, U3, axis=Height, return_inverse=True)

    assert jnp.all((values[{"U3": inverse}] == named1).array)

    #     **Returning counts**
    #  If you set ``return_counts=True``, then ``unique`` returns the number of occurrences
    #     within the input for every unique value:
    #
    #     >>> x = jnp.array([3, 4, 1, 3, 1])
    #     >>> values, counts = jnp.unique(x, return_counts=True)
    #     >>> print(values)
    #     [1 3 4]
    #     >>> print(counts)
    #     [2 2 1]
    #
    #     For multi-dimensional arrays, this also returns a 1D array of counts
    #     indicating number of occurrences along the specified axis:
    #
    #     >>> values, counts = jnp.unique(M, axis=0, return_counts=True)
    #     >>> print(values)
    #     [[1 2]
    #      [2 3]]
    #     >>> print(counts)
    #     [2 1]

    values, counts = hax.unique(x, U3, return_counts=True)

    assert jnp.all(jnp.equal(values.array, jnp.array([1, 3, 4])))
    assert jnp.all(jnp.equal(counts.array, jnp.array([2, 2, 1])))

    values, counts = hax.unique(named1, U2, axis=Height, return_counts=True)

    assert jnp.all(jnp.equal(values.array, jnp.array([[1, 2], [2, 3]])))
    assert jnp.all(jnp.equal(counts.array, jnp.array([2, 1])))


def test_unique_shortcuts():
    Height = Axis("Height", 3)
    Width = Axis("Width", 2)

    arr2d = hax.named([[1, 2], [2, 3], [1, 2]], (Height, Width))
    U = Axis("U", 3)

    # unique_values
    uv = hax.unique_values(arr2d, U)
    uv_expected = hax.unique(arr2d, U)
    assert jnp.all(uv.array == uv_expected.array)

    # unique_counts
    vc, cc = hax.unique_counts(arr2d, U)
    vc_exp, cc_exp = hax.unique(arr2d, U, return_counts=True)
    assert jnp.all(vc.array == vc_exp.array)
    assert jnp.all(cc.array == cc_exp.array)

    # unique_inverse
    Height1 = Axis("Height1", 5)
    arr1d = hax.named([3, 4, 1, 3, 1], (Height1,))
    U2 = Axis("U2", 3)
    vi, ii = hax.unique_inverse(arr1d, U2)
    vi_exp, ii_exp = hax.unique(arr1d, U2, return_inverse=True)
    assert jnp.all(vi.array == vi_exp.array)
    assert jnp.all(ii.array == ii_exp.array)

    # unique_all
    U3 = Axis("U3", 2)
    va, ia, ina, ca = hax.unique_all(arr2d, U3, axis=Height)
    va_exp, ia_exp, ina_exp, ca_exp = typing.cast(
        tuple[NamedArray, NamedArray, NamedArray, NamedArray],
        hax.unique(
            arr2d,
            U3,
            axis=Height,
            return_index=True,
            return_inverse=True,
            return_counts=True,
        ),
    )
    assert jnp.all(va.array == va_exp.array)
    assert jnp.all(ia.array == ia_exp.array)
    assert jnp.all(ina.array == ina_exp.array)
    assert jnp.all(ca.array == ca_exp.array)


def test_bincount():
    X = Axis("X", 6)
    x = hax.named([0, 1, 1, 2, 3, 1], (X,))
    B = Axis("B", 5)

    out = hax.bincount(x, B)
    expected = jnp.bincount(x.array, length=B.size)
    assert out.axes == (B,)
    assert jnp.all(out.array == expected)

    w = hax.arange((X,), dtype=jnp.float32)
    out_w = hax.bincount(x, B, weights=w)
    expected_w = jnp.bincount(x.array, weights=w.array, length=B.size)
    assert jnp.allclose(out_w.array, expected_w)


def test_allclose_array_equal_equiv():
    A = Axis("A", 2)
    B = Axis("B", 3)
    x = hax.random.uniform(PRNGKey(0), (A, B))
    y = x + 1e-6

    assert hax.allclose(x, y)
    assert not hax.allclose(x, x + 1.0)

    x1 = hax.ones((A, B))
    y_reordered = x1.rearrange((B, A))
    assert hax.array_equal(x1, y_reordered)

    scalar = hax.ones(())
    assert hax.array_equiv(x1, scalar)
    assert not hax.array_equal(x1, scalar)

    y_vec = hax.ones((B,))
    assert hax.array_equiv(x1, y_vec)
    assert not hax.array_equal(x1, y_vec)

    C = Axis("C", 4)
    z = hax.ones((C,))
    assert not hax.array_equiv(x1, z)


def test_roll_scalar_named_shift():
    H = Axis("H", 4)
    W = Axis("W", 3)

    arr = hax.arange((H, W))
    shift = hax.named(jnp.array(1), ())

    rolled = hax.roll(arr, shift, H)
    expected = jnp.roll(arr.array, shift.array, axis=0)

    assert rolled.axes == arr.axes
    assert jnp.all(rolled.array == expected)


def test_roll_bad_named_shift():
    H = Axis("H", 4)
    W = Axis("W", 3)

    arr = hax.arange((H, W))
    shift = hax.arange((Axis("dummy", 2),))

    with pytest.raises(TypeError):
        hax.roll(arr, shift, H)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, MutableMapping, Sequence, TypeAlias, TypeVar

T = TypeVar("T")
U = TypeVar("U")
py_slice = slice
slice_t: TypeAlias = slice


def index_where(pred: Callable[[T], bool], xs: Sequence[T], start: int = 0) -> int:
    for i in range(start, len(xs)):
        if pred(xs[i]):
            return i
    raise ValueError("No element satisfies predicate")


class IdentityMap(MutableMapping[T, U]):
    """Map that compares keys by identity.

    This is a map that compares keys by identity instead of equality. It is
    useful for storing objects that are not hashable or that should be compared
    by identity.

    This is a mutable mapping, but it does not support the ``__hash__`` method
    and therefore cannot be used as a dictionary key or as an element of another
    set.
    """

    def __init__(self, iterable=None):
        self._data = {}
        if iterable is not None:
            self.update(iterable)

    def __contains__(self, key):
        return id(key) in self._data

    def __getitem__(self, key):
        return self._data[id(key)][1]

    def __setitem__(self, key, value):
        self._data[id(key)] = [key, value]

    def __delitem__(self, key):
        del self._data[id(key)]

    def __iter__(self):
        return (x[0] for x in self._data.values())

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"IdentityMap({list(repr(x) for x in self._data.values())})"

    def __str__(self):
        return f"IdentityMap({list(str(x) for x in self._data.values())})"



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from typing import Callable

import equinox as eqx
import jax.nn
import jax.random as jrandom
import pytest
from jax import numpy as jnp

import haliax as hax
from haliax import Axis, NamedArray


def _compare_eqx_and_haliax(hax_mod: eqx.Module, eqx_mod: eqx.Module):
    def f(x: NamedArray, *args, **kwargs):
        unnamed_x = x.array
        hax_out = hax_mod(x, *args, **kwargs)  # type: ignore
        eqx_out = eqx_mod(unnamed_x, *args, **kwargs)  # type: ignore

        assert jnp.allclose(hax_out.array, eqx_out)
        return hax_out

    return f


def test_layer_norm():
    H = Axis("H", 10)
    hax_ln = hax.nn.LayerNorm.init(H)
    eqx_ln = eqx.nn.LayerNorm(shape=(H.size,))

    f = _compare_eqx_and_haliax(hax_ln, eqx_ln)
    out = f(hax.random.uniform(jrandom.PRNGKey(0), (H,)))

    assert out.axes == (H,)


def test_dropout():
    H = Axis("H", 10)
    key = jrandom.PRNGKey(0)
    hax_dropout = hax.nn.Dropout(0.5)
    eqx_dropout = eqx.nn.Dropout(0.5)

    f = _compare_eqx_and_haliax(hax_dropout, eqx_dropout)
    out = f(hax.random.uniform(jrandom.PRNGKey(0), (H,)), key=key, inference=False)

    assert out.axes == (H,)


def test_one_hot():
    i, c = hax.make_axes(i=3, c=3)
    actual = hax.nn.one_hot(hax.NamedArray(jnp.array([0, 1, 2]), (i,)), c)
    expected = jnp.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

    assert actual.axes == (i, c)
    assert jnp.all(jnp.isclose(actual.array, expected))

    actual = hax.nn.one_hot(hax.NamedArray(jnp.array([1, 2, 0]), (i,)), c)
    expected = jnp.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0], [1.0, 0.0, 0.0]])
    assert actual.axes == (i, c)
    assert jnp.all(jnp.isclose(actual.array, expected))


def test_one_hot_scalar_array():
    (c,) = hax.make_axes(c=3)

    scalar_value = jnp.array(1, dtype=jnp.int32)
    actual = hax.nn.one_hot(scalar_value, c)
    expected = jnp.array([0.0, 1.0, 0.0])

    assert actual.axes == (c,)
    assert jnp.all(jnp.isclose(actual.array, expected))

    negative_scalar = jnp.array(-1, dtype=jnp.int32)
    actual = hax.nn.one_hot(negative_scalar, c)
    expected = jnp.array([0.0, 0.0, 1.0])

    assert actual.axes == (c,)
    assert jnp.all(jnp.isclose(actual.array, expected))


def test_one_hot_out_of_bound():
    i, c = hax.make_axes(i=2, c=3)
    actual = hax.nn.one_hot(hax.NamedArray(jnp.array([-1, 3]), (i,)), c)
    expected = jnp.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    assert jnp.all(jnp.isclose(actual.array, expected))


def test_standardize():
    b, c = hax.make_axes(b=2, c=3)
    actual = hax.nn.standardize(hax.NamedArray(jnp.array([0, 1, 2]), (c,)), c)
    expected = jax.nn.standardize(jnp.array([0, 1, 2]), axis=0)

    assert actual.axes == (c,)
    assert jnp.all(jnp.isclose(actual.array, expected))

    actual = hax.nn.standardize(hax.NamedArray(jnp.array([[0, 1, 2], [3, 4, 5]]), (b, c)), c)
    expected = jax.nn.standardize(jnp.array([[0, 1, 2], [3, 4, 5]]), axis=1)

    assert actual.axes == (b, c)
    assert jnp.all(jnp.isclose(actual.array, expected))

    actual = hax.nn.standardize(hax.NamedArray(jnp.array([[0, 1, 2], [3, 4, 5]]), (b, c)), b)
    expected = jax.nn.standardize(jnp.array([[0, 1, 2], [3, 4, 5]]), axis=0)

    assert actual.axes == (b, c)
    assert jnp.all(jnp.isclose(actual.array, expected))

    # test passing in where
    mask = hax.NamedArray(jnp.array([True, False, True]), (c,))
    actual = hax.nn.standardize(hax.NamedArray(jnp.array([[0, 1, 2], [3, 4, 5]]), (b, c)), b, where=mask)
    expected = jax.nn.standardize(jnp.array([[0, 1, 2], [3, 4, 5]]), axis=0, where=mask.array)

    assert actual.axes == (b, c)
    assert jnp.all(jnp.isclose(actual.array, expected) | jnp.isnan(expected))

    # now mean/variance
    input = hax.NamedArray(jnp.array([[0, 1, 2], [3, 4, 5]]), (b, c))
    mean = hax.mean(input, c)
    variance = hax.var(input, c)
    actual = hax.nn.standardize(input, c, mean=mean, variance=variance)
    expected = jax.nn.standardize(
        input.array, axis=1, mean=mean.array.reshape(-1, 1), variance=variance.array.reshape(-1, 1)
    )

    assert actual.axes == (b, c)
    assert jnp.all(jnp.isclose(actual.array, expected))


@pytest.mark.parametrize("depth", [0, 1, 2, 3, 4, 5])
def test_mlp(depth):
    key = jrandom.PRNGKey(0)
    H, C, W, E = hax.make_axes(H=10, C=12, W=14, E=16)

    hax_mlp = hax.nn.MLP.init((H, C, W), E, width=8, depth=depth, key=key)
    x = hax.random.uniform(key, (H, C, W))
    assert hax_mlp(x).axes == (E,)

    hax_mlp = hax.nn.MLP.init((H, W), E, width=8, depth=depth, key=key)
    assert hax_mlp(x).axes == (C, E)

    # with a named width
    M = Axis("M", 18)
    hax_mlp = hax.nn.MLP.init((H, W), E, width=M, depth=depth, key=key)
    assert hax_mlp(x).axes == (C, E)

    # ensure that we actually use the name for the named width
    if depth > 0:
        assert hax_mlp.layers[0].Out == M

        if depth % 2 == 0:
            assert hax_mlp.layers[-1].In == M.alias("M2")
        else:
            assert hax_mlp.layers[-1].In == M

        for i in range(1, depth):
            if i % 2 == 0:
                assert hax_mlp.layers[i].In == M.alias("M2")
                assert hax_mlp.layers[i].Out == M
            else:
                assert hax_mlp.layers[i].In == M
                assert hax_mlp.layers[i].Out == M.alias("M2")


def test_linear_has_no_function_leaves_by_default():
    H, C, W, E = hax.make_axes(H=10, C=12, W=14, E=16)

    hax_linear = hax.nn.Linear.init((H, C, W), E, key=jrandom.PRNGKey(0))
    assert all(not isinstance(v, Callable) for v in jax.tree_util.tree_leaves(hax_linear))  # type: ignore


@pytest.mark.parametrize(
    "input_data, axes",
    [
        (jnp.array([-2.0, -1.0, 0.0, 1.0, 2.0]), (hax.Axis("X", 5),)),
        (jnp.array([[1.0, -1.0], [0.0, 2.0]]), (hax.Axis("Y", 2), hax.Axis("Z", 2))),
        (jnp.array([jnp.nan, 1.0, -1.0]), (hax.Axis("A", 3),)),
        (jnp.array([jnp.inf, -jnp.inf, 0.0]), (hax.Axis("B", 3),)),
    ],
)
@pytest.mark.parametrize("dtype", [jnp.float16, jnp.float32, jnp.bfloat16])
@pytest.mark.parametrize("use_jit", [False, True])
def test_relu_squared_robust(input_data, axes, dtype, use_jit):
    input_data = input_data.astype(dtype)
    x = hax.named(input_data, axes)

    # Manually compute the expected output using the base JAX functions
    expected_raw = jnp.square(jax.nn.relu(input_data))
    expected = hax.named(expected_raw, axes)

    f = hax.nn.relu_squared
    if use_jit:
        f = hax.named_jit(f)

    # Apply the relu_squared function
    actual = f(x)

    # Check that the output is a NamedArray with the correct axes and dtype
    assert isinstance(actual, hax.NamedArray)
    assert actual.axes == expected.axes
    assert actual.dtype == expected.dtype

    # Check that the values are correct, handling NaNs correctly
    assert jnp.allclose(actual.array, expected.array, equal_nan=True)


@pytest.mark.parametrize("use_jit", [False, True])
def test_relu_squared_scalar(use_jit):
    f = hax.nn.relu_squared
    if use_jit:
        f = jax.jit(f)

    x = 5.0
    expected = 25.0
    actual = f(x)
    assert jnp.allclose(actual, expected)

    x_neg = -5.0
    expected_neg = 0.0
    actual_neg = f(x_neg)
    assert jnp.allclose(actual_neg, expected_neg)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


from typing import Any, Callable, Literal, Protocol, TypeAlias

import jax.numpy as jnp
import numpy as np
from jax.lax import Precision
from jaxtyping import PyTree

DType: TypeAlias = np.dtype

try:
    from jax.typing import DTypeLike
except ImportError:
    # Cribbed from jax.typing, for older versions of JAX
    class SupportsDType(Protocol):
        @property
        def dtype(self) -> DType: ...

    DTypeLike = (
        str  # like 'float32', 'int32'
        | type  # like np.float32, np.int32, float, int
        | np.dtype  # like np.dtype('float32'), np.dtype('int32')
        | SupportsDType  # like jnp.float32, jnp.int32
    )


Scalar = float | int | jnp.ndarray  # ndarray b/c array(1) is a scalar
IntScalar = int | jnp.ndarray

PrecisionLike = None | str | Precision | tuple[str, str] | tuple[Precision, Precision]

GatherScatterModeStr = Literal["promise_in_bounds", "clip", "drop", "fill"]


FilterSpec = bool | Callable[[Any], bool]
"""
A filter specification. Typically used on a pytree to filter out certain subtrees. Boolean values are
treated as-is, while callables are called on each element of the pytree. If the callable returns True, the element
is kept, otherwise it is filtered out.
"""

FilterTree = FilterSpec | PyTree[FilterSpec]



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0

import math
from abc import ABC, abstractmethod
from dataclasses import dataclass

import haliax as hax
import equinox as eqx

from ..axis import AxisSpec


class AbstractReparam(ABC):
    """Abstract base class for abc-parameterization rules.

    Defines the interface for active scaling of parameters (a),
    computing initialization scales (b), and learning rate scaling (c)

    See: https://arxiv.org/abs/2011.14522
    """

    @staticmethod
    @abstractmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        """Return the scaling factor for initializing weights
        given input and output axes."""
        raise NotImplementedError

    @property
    @abstractmethod
    def lr_scale(self):
        """Return the learning-rate scaling factor."""
        raise NotImplementedError

    @property
    @abstractmethod
    def active_scale(self):
        """Return the scaling applied to activations."""
        raise NotImplementedError


@dataclass
class AbstractLinearReparam(AbstractReparam):
    """Base class for linear-layer reparameterizations.

    Stores input and output axis specifications, and inherits
    the reparameterization interface.
    """

    In: AxisSpec
    Out: AxisSpec


class LinearStandardParam(AbstractLinearReparam):
    """Standard (non-muP) parameterization for linear layers.

    Uses the usual fan-in scaling for initialization and
    leaves learning rate and activation scaling unchanged.
    """

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1 / math.sqrt(hax.axis_size(In))

    @property
    def active_scale(self):
        return 1

    @property
    def lr_scale(self):
        return 1


class InputLinearMup(AbstractLinearReparam):
    """muP-style parameterization for input linear layers.

    Uses no scaling on initialization or learning rate.
    See: https://arxiv.org/abs/2011.14522 (Maximal Update Parametrization)
    """

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1

    @property
    def active_scale(self):
        return 1

    @property
    def lr_scale(self):
        return 1


class HiddenLinearMup(AbstractLinearReparam):
    """muP-style parameterization for hidden linear layers.

    Applies fan-in scaling at initialization and scales
    learning rate inversely with layer width.
    """

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1 / math.sqrt(hax.axis_size(In))

    @property
    def active_scale(self):
        return 1

    @property
    def lr_scale(self):
        return 1 / hax.axis_size(self.In)


class OutputLinearMup(AbstractLinearReparam):
    """muP-style parameterization for output linear layers.

    Uses unit initialization and applies inverse-width
    scaling to the output activations.
    """

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1

    @property
    def active_scale(self):
        return 1 / hax.axis_size(self.In)

    @property
    def lr_scale(self):
        return 1


@dataclass
class AbstractEmbeddingReparam(AbstractReparam):
    """Base class for embedding-layer reparameterizations.

    Defines the interface for both embedding and unembedding
    scaling rules.
    """

    Embed: AxisSpec
    Vocab: AxisSpec

    @property
    @abstractmethod
    def unembed_active_scale(self):
        """Scaling factor applied when unembedding embeddings."""
        raise NotImplementedError


class EmbeddingStandardParam(AbstractEmbeddingReparam):
    """Standard embedding parameterization."""

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1 / hax.axis_size(Out)

    @property
    def active_scale(self):
        return 1

    @property
    def lr_scale(self):
        return 1

    @property
    def unembed_active_scale(self):
        return 1


class EmbeddingMup(AbstractEmbeddingReparam):
    """muP-style parameterization for embeddings.

    Keeps initialization and learning-rate scaling neutral,
    but applies inverse-width scaling to unembedding outputs for tied weights.
    See: https://www.cerebras.ai/blog/the-practitioners-guide-to-the-maximal-update-parameterization
    """

    @staticmethod
    def init_scale(In: AxisSpec, Out: AxisSpec):
        return 1

    @property
    def active_scale(self):
        return 1

    @property
    def lr_scale(self):
        return 1

    @property
    def unembed_active_scale(self):
        return 1 / hax.axis_size(self.Embed)


class ReparamEnabled(ABC):
    """Mixin for modules that support reparameterization.

    Stores an abstract `reparam` attribute that specifies
    how initialization and scaling are handled.
    """

    reparam: eqx.AbstractVar[AbstractReparam]



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


# This whole file is copied from Equinox.
# (c) 2023, Google LLC. and/or Patrick Kidger. Apache 2.0 licensed.
# Patrick doesn't like that I depend on Equinox internals, so I copied this stuff
import functools as ft
import types
import warnings
import weakref
from typing import Any

import jax.tree_util as jtu


def _strip_wrapped_partial(fun):
    if hasattr(fun, "__wrapped__"):  # ft.wraps
        return _strip_wrapped_partial(fun.__wrapped__)
    if isinstance(fun, ft.partial):
        return _strip_wrapped_partial(fun.func)
    return fun


internal_caches = []  # type: ignore
internal_lru_caches = []  # type: ignore


def clear_caches():
    """Clears internal Equinox caches.

    Best used before calling `jax.clear_caches()` or `jax.clear_backends()`.

    **Arguments:**

    None.

    **Returns:**

    None.
    """
    for cache in internal_caches:
        cache.clear()
    for cache in internal_lru_caches:
        cache.cache_clear()


def _default_cache_key():
    assert False


def compile_cache(cacheable_fn):
    cache = weakref.WeakKeyDictionary()  # type: ignore
    internal_caches.append(cache)

    def cached_fn_impl(leaves, treedef):
        user_fn_names, args, kwargs = jtu.tree_unflatten(treedef, leaves)
        return cacheable_fn(user_fn_names, *args, **kwargs)

    @ft.wraps(cacheable_fn)
    def wrapped_cacheable_fn(user_fn, *args, **kwargs):
        user_fn = _strip_wrapped_partial(user_fn)
        # Best-effort attempt to clear the cache of old and unused entries.
        cache_key: Any
        if type(user_fn) is types.FunctionType:  # noqa: E721
            cache_key = user_fn
        else:
            cache_key = _default_cache_key

        try:
            user_fn_names = user_fn.__name__, user_fn.__qualname__
        except AttributeError:
            user_fn_names = type(user_fn).__name__, type(user_fn).__qualname__
        leaves, treedef = jtu.tree_flatten((user_fn_names, args, kwargs))
        leaves = tuple(leaves)

        try:
            cached_fn = cache[cache_key]
        except KeyError:
            cached_fn = cache[cache_key] = ft.lru_cache(maxsize=None)(cached_fn_impl)
        return cached_fn(leaves, treedef)

    def delete(user_fn):
        user_fn = _strip_wrapped_partial(user_fn)
        if type(user_fn) is types.FunctionType:  # noqa: E721
            try:
                del cache[user_fn]
            except KeyError:
                warnings.warn(f"Could not delete cache for function {user_fn}. Has it already been deleted?")
        else:
            warnings.warn("Could not delete non-function from cache.")

    wrapped_cacheable_fn.delete = delete  # type: ignore
    return wrapped_cacheable_fn



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import functools

# Removed unused import

import equinox as eqx
import jax
import jax.tree_util as jtu
from jaxtyping import PRNGKeyArray, PyTree

import haliax.nn

from .axis import AxisSelector
from .core import NamedArray
from .jax_utils import maybe_rng_split
from .util import is_named_array


def tree_map(fn, tree, *rest, is_leaf=None):
    """
    Version of [jax.tree_util.tree_map][] that automatically treats NamedArrays as leaves.
    """
    old_is_leaf = is_leaf
    if is_leaf is None:
        is_leaf = lambda x: isinstance(x, NamedArray)
    else:
        is_leaf = lambda x: old_is_leaf(x) or is_named_array(x)

    return jax.tree.map(fn, tree, *rest, is_leaf=is_leaf)


def scan_aware_tree_map(fn, tree, *rest, is_leaf=None):
    """
    Version of [haliax.tree_util.tree_map][] that is aware of the scan-layer pattern, specifically as implemented
    in hax.nn.Stacked. This function will (implicitly) apply the transform to each layer in each Stacked module
    (using vmap). If there are no Stacked modules in the tree, this function is equivalent to [haliax.tree_util.tree_map][].

    """
    old_is_leaf = is_leaf
    if is_leaf is None:
        is_leaf = lambda x: isinstance(x, haliax.nn.Stacked)
    else:
        is_leaf = lambda x: old_is_leaf(x) or isinstance(x, haliax.nn.Stacked)

    mapped_fn = functools.partial(scan_aware_tree_map, fn, is_leaf=is_leaf)

    def rec_fn(x, *rest):
        if isinstance(x, haliax.nn.Stacked):
            new_inner = haliax.vmap(mapped_fn, x.Block)(x.stacked, *[r.stacked for r in rest])
            return dataclasses.replace(x, stacked=new_inner)  # type: ignore
        else:
            return fn(x, *rest)

    return tree_map(rec_fn, tree, *rest, is_leaf=is_leaf)


def tree_flatten(tree, is_leaf=None):
    """
    Version of [jax.tree_util.tree_flatten][] that automatically treats NamedArrays as leaves.
    """
    if is_leaf is None:
        is_leaf = lambda x: isinstance(x, NamedArray)
    else:
        is_leaf = lambda x: is_leaf(x) or is_named_array(x)

    return jax.tree_util.tree_flatten(tree, is_leaf=is_leaf)


def tree_unflatten(treedef, leaves):
    """
    Provided for consistency with tree_flatten.
    """
    return jax.tree_util.tree_unflatten(treedef, leaves)


def tree_leaves(tree, is_leaf=None):
    """
    Version of [jax.tree_util.tree_leaves][] that automatically treats NamedArrays as leaves.
    """
    if is_leaf is None:
        is_leaf = lambda x: isinstance(x, NamedArray)
    else:
        is_leaf = lambda x: is_leaf(x) or is_named_array(x)

    return jax.tree_util.tree_leaves(tree, is_leaf=is_leaf)


def tree_structure(tree, is_leaf=None):
    """
    Version of [jax.tree_util.tree_structure][] that automatically treats NamedArrays as leaves.
    """
    if is_leaf is None:
        is_leaf = lambda x: isinstance(x, NamedArray)
    else:
        is_leaf = lambda x: is_leaf(x) or is_named_array(x)

    return jax.tree_util.tree_structure(tree, is_leaf=is_leaf)


def resize_axis(tree: PyTree[NamedArray], old_axis: AxisSelector, new_size: int, key: PRNGKeyArray | None = None):
    """Resizes the NamedArrays of a PyTree along a given axis. If the array needs to grow and key is not none, then the
    new elements are sampled from a truncated normal distribution with the same mean and standard deviation as the
    existing elements. If the key is none, they're just initialized to the mean. If the array needs to shrink, then it's
    truncated.

    Note: if you have a module that stores a reference to the old axis, then you'll need to update that reference
    manually.

    """
    import haliax.random

    def _resize_one(x, key):
        if not is_named_array(x):
            return x

        assert isinstance(x, NamedArray)

        try:
            current_axis = x.resolve_axis(old_axis)
        except ValueError:
            return x

        if new_size == current_axis.size:
            return x
        elif current_axis.size > new_size:
            return x.slice(current_axis, start=0, length=new_size)
        else:
            num_padding = new_size - current_axis.size

            mean = x.mean(current_axis)
            std = x.std(current_axis)

            # the shape of the padding is the same as the original array, except with the axis size changed
            padding_axes = list(x.axes)
            padding_axes[padding_axes.index(current_axis)] = current_axis.resize(num_padding)

            if key is None:
                padding = mean.broadcast_axis(padding_axes)
            else:
                padding = haliax.random.truncated_normal(key, padding_axes, lower=-2, upper=2) * std + mean

            return haliax.concatenate(current_axis.name, [x, padding])

    leaves, structure = jax.tree_util.tree_flatten(tree, is_leaf=is_named_array)
    keys = maybe_rng_split(key, len(leaves))

    new_leaves = [_resize_one(x, key) for x, key in zip(leaves, keys)]

    return jax.tree_util.tree_unflatten(structure, new_leaves)


# old version of eqx's partition functions
def hashable_partition(pytree, filter_spec):
    dynamic, static = eqx.partition(pytree, filter_spec)
    static_leaves, static_treedef = jtu.tree_flatten(static)
    static_leaves = tuple(static_leaves)
    return dynamic, (static_leaves, static_treedef)


def hashable_combine(dynamic, static):
    static_leaves, static_treedef = static
    static = jtu.tree_unflatten(static_treedef, static_leaves)
    return eqx.combine(dynamic, static)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


import equinox as eqx
import jax
import jax.numpy as jnp

import haliax as hax
from haliax.nn.conv import Conv, ConvTranspose


def test_conv_basic_equiv_to_eqx():
    key = jax.random.PRNGKey(0)
    In = hax.Axis("In", 3)
    Out = hax.Axis("Out", 4)
    hax_conv = Conv.init(("Height", "Width"), In, Out, kernel_size=3, key=key)
    eqx_conv = eqx.nn.Conv(2, 3, 4, kernel_size=3, key=key)

    assert hax_conv.weight.array.shape == eqx_conv.weight.shape
    assert hax_conv.bias.array.shape == eqx_conv.bias.shape[0:1]
    assert jnp.all(hax_conv.weight.array == eqx_conv.weight)

    input = hax.random.normal(jax.random.PRNGKey(1), (In, hax.Axis("Height", 5), hax.Axis("Width", 6)))
    hax_output = hax_conv(input)
    eqx_output = eqx_conv(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test batched
    input = hax.random.normal(
        jax.random.PRNGKey(1), (hax.Axis("Batch", 2), In, hax.Axis("Height", 5), hax.Axis("Width", 6))
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx_conv)(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test multibatch
    input = hax.random.normal(
        jax.random.PRNGKey(1),
        (hax.Axis("Batch", 2), hax.Axis("Batch2", 3), In, hax.Axis("Height", 5), hax.Axis("Width", 6)),
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx.filter_vmap(eqx_conv))(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.allclose(hax_output.array, eqx_output)

    input = hax.random.normal(
        jax.random.PRNGKey(1),
        (
            hax.Axis("Batch", 2),
            In,
            hax.Axis("Height", 5),
            hax.Axis("Width", 6),
            hax.Axis("Batch2", 3),
        ),
    )
    hax_output = hax_conv(input).rearrange(("Batch", "Batch2", "Out", "Height", "Width"))
    eqx_output = eqx.filter_vmap(eqx.filter_vmap(eqx_conv))(
        input.rearrange(("Batch", "Batch2", "In", "Height", "Width")).array
    )

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.allclose(hax_output.array, eqx_output)


def test_conv_grouped_equiv_to_eqx():
    key = jax.random.PRNGKey(0)
    In = hax.Axis("In", 4)
    Out = hax.Axis("Out", 6)
    hax_conv = Conv.init(("Height", "Width"), In, Out, kernel_size=3, groups=2, key=key)
    eqx_conv = eqx.nn.Conv(2, 4, 6, kernel_size=3, groups=2, key=key)

    assert hax_conv.weight.array.shape == eqx_conv.weight.shape
    assert hax_conv.bias.array.shape == eqx_conv.bias.shape[0:1]
    assert jnp.all(hax_conv.weight.array == eqx_conv.weight)

    input = hax.random.normal(jax.random.PRNGKey(1), (In, hax.Axis("Height", 5), hax.Axis("Width", 6)))
    eqx_output = eqx_conv(input.array)
    hax_output = hax_conv(input)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test batched
    input = hax.random.normal(
        jax.random.PRNGKey(1), (hax.Axis("Batch", 2), In, hax.Axis("Height", 5), hax.Axis("Width", 6))
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx_conv)(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test multibatch
    input = hax.random.normal(
        jax.random.PRNGKey(1),
        (hax.Axis("Batch", 2), hax.Axis("Batch2", 3), In, hax.Axis("Height", 5), hax.Axis("Width", 6)),
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx.filter_vmap(eqx_conv))(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.allclose(hax_output.array, eqx_output)


def test_conv_weird_order():
    key = jax.random.PRNGKey(0)
    In = hax.Axis("In", 3)
    Out = hax.Axis("Out", 4)
    hax_conv = Conv.init(("Height", "Width"), In, Out, kernel_size=3, key=key)
    eqx_conv = eqx.nn.Conv(2, 3, 4, kernel_size=3, key=key)

    assert hax_conv.weight.array.shape == eqx_conv.weight.shape
    assert hax_conv.bias.array.shape == eqx_conv.bias.shape[0:1]
    assert jnp.all(hax_conv.weight.array == eqx_conv.weight)

    input = hax.random.normal(
        jax.random.PRNGKey(1), (hax.Axis("Batch", 2), In, hax.Axis("Height", 5), hax.Axis("Width", 6))
    )
    hax_output = hax_conv(input)

    # test weird orders
    input = input.rearrange(("In", "Height", "Width", "Batch"))
    hax_output2 = hax_conv(input).rearrange(("Batch", "Out", "Height", "Width"))

    assert jnp.allclose(hax_output.array, hax_output2.array)


def test_conv_transpose_basic_equiv_to_eqx():
    key = jax.random.PRNGKey(0)
    In = hax.Axis("In", 3)
    Out = hax.Axis("Out", 4)
    hax_conv = ConvTranspose.init(
        ("Height", "Width"), In, Out, kernel_size=3, dilation=2, output_padding=1, stride=2, key=key
    )
    eqx_conv = eqx.nn.ConvTranspose(2, 3, 4, kernel_size=3, dilation=2, output_padding=1, stride=2, key=key)

    assert hax_conv.weight.array.shape == eqx_conv.weight.shape
    assert hax_conv.bias.array.shape == eqx_conv.bias.shape[0:1]
    assert jnp.all(hax_conv.weight.array == eqx_conv.weight)

    input = hax.random.normal(jax.random.PRNGKey(1), (In, hax.Axis("Height", 5), hax.Axis("Width", 6)))
    hax_output = hax_conv(input)
    eqx_output = eqx_conv(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test batched
    input = hax.random.normal(
        jax.random.PRNGKey(1), (hax.Axis("Batch", 2), In, hax.Axis("Height", 5), hax.Axis("Width", 6))
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx_conv)(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.all(hax_output.array == eqx_output)

    # test multibatch
    input = hax.random.normal(
        jax.random.PRNGKey(1),
        (hax.Axis("Batch", 2), hax.Axis("Batch2", 3), In, hax.Axis("Height", 5), hax.Axis("Width", 6)),
    )
    hax_output = hax_conv(input)
    eqx_output = eqx.filter_vmap(eqx.filter_vmap(eqx_conv))(input.array)

    assert hax_output.array.shape == eqx_output.shape
    assert jnp.allclose(hax_output.array, eqx_output)


def test_weird_orders_conv_transpose():
    key = jax.random.PRNGKey(0)
    In = hax.Axis("In", 3)
    Out = hax.Axis("Out", 4)
    hax_conv = ConvTranspose.init(
        ("Height", "Width"), In, Out, kernel_size=3, dilation=2, output_padding=1, stride=2, key=key
    )

    input = hax.random.normal(
        jax.random.PRNGKey(1), (hax.Axis("Batch", 2), In, hax.Axis("Height", 5), hax.Axis("Width", 6))
    )
    hax_output = hax_conv(input)

    # test weird orders
    input = input.rearrange(("In", "Height", "Width", "Batch"))
    hax_output2 = hax_conv(input).rearrange(("Batch", "Out", "Height", "Width"))

    assert jnp.allclose(hax_output.array, hax_output2.array)



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0

"""Named wrappers around :mod:`jax.numpy.fft`.

These functions mirror the behaviour of their :mod:`jax.numpy.fft` counterparts
while accepting named axes.  Instead of separate ``fftn``/``fft2`` variants we
provide a single ``fft`` family of functions whose ``axis`` argument controls
which axes are transformed.

The ``axis`` parameter can be one of:

* ``None`` – operate on the last axis.
* ``str`` – name of an existing axis in the input.
* :class:`~haliax.Axis` – specifies both the axis to transform (by name) and the
  desired FFT length.  The output axis is replaced by the provided ``Axis``.
* ``dict`` – mapping from axis selectors (names or ``Axis`` objects) to optional
  sizes.  A value of ``None`` uses the existing axis length.  The mapping order
  determines the order of transforms and dispatches to the ``n``‑dimensional
  variants in :mod:`jax.numpy.fft`.

Example
-------

```python
X, Y = hax.make_axes(X=4, Y=6)
arr = hax.arange((X, Y))

# 1D transform along ``Y``
hax.fft(arr, axis="Y")

# 2D transform across both axes
hax.fft(arr, axis={"X": None, "Y": None})

# Resize the ``Y`` axis before transforming
hax.fft(arr, axis={"Y": Axis("Y", 8)})
```
"""

from __future__ import annotations

from typing import Mapping, MutableSequence, Sequence

import jax.numpy.fft as jfft

from .axis import Axis, AxisSelector, AxisSelection
from .core import NamedArray

AxisSizeLike = int | Axis | None
AxisMapping = Mapping[AxisSelector, AxisSizeLike]


def _single_axis(a: NamedArray, axis: AxisSelector | None):
    if axis is None:
        idx = a.ndim - 1
        ax = a.axes[idx]
        n = None
    elif isinstance(axis, Axis):
        idx = a.axis_indices(axis.name)
        if idx is None:
            raise ValueError(f"Axis {axis} not found in {a.axes}")
        ax = axis
        n = axis.size
    else:
        idx = a.axis_indices(axis)
        if idx is None:
            raise ValueError(f"Axis {axis} not found in {a.axes}")
        ax = a.axes[idx]
        n = None
    return idx, ax, n


def _multi_axis(a: NamedArray, axis: AxisMapping):
    axes_idx: MutableSequence[int] = []
    sizes: MutableSequence[int] = []
    new_axes = list(a.axes)
    for key, val in axis.items():
        idx = a.axis_indices(key)
        if idx is None:
            raise ValueError(f"Axis {key} not found in {a.axes}")
        ax = a.axes[idx]
        if isinstance(val, Axis):
            size = val.size
            new_axes[idx] = val
        elif val is None:
            size = ax.size
        else:
            size = int(val)
            new_axes[idx] = ax.resize(size)
        axes_idx.append(idx)
        sizes.append(size)
    return list(axes_idx), list(sizes), new_axes


def fft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.fft`.

    See module level documentation for the behaviour of the ``axis`` argument.
    """

    if isinstance(axis, Mapping):
        axes_idx, sizes, new_axes = _multi_axis(a, axis)
        out = jfft.fftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    elif isinstance(axis, Sequence) and not isinstance(axis, (str, Axis)):
        axes_idx, sizes, new_axes = _multi_axis(a, {ax: None for ax in axis})
        out = jfft.fftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    else:
        idx, new_axis, n = _single_axis(a, axis)  # type: ignore[arg-type]
        out = jfft.fft(a.array, n=n, axis=idx, norm=norm)
        axes = list(a.axes)
        axes[idx] = new_axis
        return NamedArray(out, tuple(axes))


def ifft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.ifft`."""

    if isinstance(axis, Mapping):
        axes_idx, sizes, new_axes = _multi_axis(a, axis)
        out = jfft.ifftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    elif isinstance(axis, Sequence) and not isinstance(axis, (str, Axis)):
        axes_idx, sizes, new_axes = _multi_axis(a, {ax: None for ax in axis})
        out = jfft.ifftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    else:
        idx, new_axis, n = _single_axis(a, axis)  # type: ignore[arg-type]
        out = jfft.ifft(a.array, n=n, axis=idx, norm=norm)
        axes = list(a.axes)
        axes[idx] = new_axis
        return NamedArray(out, tuple(axes))


def rfft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.rfft`."""

    if isinstance(axis, Mapping):
        axes_idx, sizes, new_axes = _multi_axis(a, axis)
        out = jfft.rfftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        last_idx = axes_idx[-1]
        last_in = sizes[-1]
        new_axes[last_idx] = new_axes[last_idx].resize(last_in // 2 + 1)
        return NamedArray(out, tuple(new_axes))
    elif isinstance(axis, Sequence) and not isinstance(axis, (str, Axis)):
        axes_idx, sizes, new_axes = _multi_axis(a, {ax: None for ax in axis})
        out = jfft.rfftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        last_idx = axes_idx[-1]
        last_in = sizes[-1]
        new_axes[last_idx] = new_axes[last_idx].resize(last_in // 2 + 1)
        return NamedArray(out, tuple(new_axes))
    else:
        idx, ax, n = _single_axis(a, axis)  # type: ignore[arg-type]
        out = jfft.rfft(a.array, n=n, axis=idx, norm=norm)
        length = n if n is not None else ax.size
        new_axis = ax.resize(length // 2 + 1)
        axes = list(a.axes)
        axes[idx] = new_axis
        return NamedArray(out, tuple(axes))


def _single_axis_irfft(a: NamedArray, axis: AxisSelector | None):
    if axis is None:
        idx = a.ndim - 1
        in_ax = a.axes[idx]
        length = (in_ax.size - 1) * 2
        out_ax = in_ax.resize(length)
        n = None
    elif isinstance(axis, Axis):
        idx = a.axis_indices(axis.name)
        if idx is None:
            raise ValueError(f"Axis {axis} not found in {a.axes}")
        out_ax = axis
        n = axis.size
    else:
        idx = a.axis_indices(axis)
        if idx is None:
            raise ValueError(f"Axis {axis} not found in {a.axes}")
        in_ax = a.axes[idx]
        length = (in_ax.size - 1) * 2
        out_ax = in_ax.resize(length)
        n = None
    return idx, out_ax, n


def _multi_axis_irfft(a: NamedArray, axis: AxisMapping):
    axes_idx: MutableSequence[int] = []
    sizes: MutableSequence[int] = []
    new_axes = list(a.axes)
    items = list(axis.items())
    for i, (key, val) in enumerate(items):
        idx = a.axis_indices(key)
        if idx is None:
            raise ValueError(f"Axis {key} not found in {a.axes}")
        ax = a.axes[idx]
        if isinstance(val, Axis):
            size = val.size
            new_axes[idx] = val
        elif val is None:
            if i == len(items) - 1:
                size = (ax.size - 1) * 2
            else:
                size = ax.size
            new_axes[idx] = ax.resize(size)
        else:
            size = int(val)
            new_axes[idx] = ax.resize(size)
        axes_idx.append(idx)
        sizes.append(size)
    return list(axes_idx), list(sizes), new_axes


def irfft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.irfft`."""

    if isinstance(axis, Mapping):
        axes_idx, sizes, new_axes = _multi_axis_irfft(a, axis)
        out = jfft.irfftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    elif isinstance(axis, Sequence) and not isinstance(axis, (str, Axis)):
        axes_idx, sizes, new_axes = _multi_axis_irfft(a, {ax: None for ax in axis})
        out = jfft.irfftn(a.array, s=tuple(sizes), axes=tuple(axes_idx), norm=norm)
        return NamedArray(out, tuple(new_axes))
    else:
        idx, out_ax, n = _single_axis_irfft(a, axis)  # type: ignore[arg-type]
        out = jfft.irfft(a.array, n=n, axis=idx, norm=norm)
        axes = list(a.axes)
        axes[idx] = out_ax
        return NamedArray(out, tuple(axes))


def hfft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.hfft`.

    Only a single axis is supported; passing a dictionary with more than one
    entry will raise an error.
    """

    if isinstance(axis, Mapping):
        if len(axis) != 1:
            raise ValueError("hfft only supports a single axis")
        key, val = next(iter(axis.items()))
        if isinstance(val, Axis):
            axis = val
        elif val is None:
            axis = key
        else:
            name = key.name if isinstance(key, Axis) else key
            axis = Axis(name, int(val))

    idx, out_ax, n = _single_axis_irfft(a, axis)  # type: ignore[arg-type]
    out = jfft.hfft(a.array, n=n, axis=idx, norm=norm)
    axes = list(a.axes)
    axes[idx] = out_ax
    return NamedArray(out, tuple(axes))


def ihfft(
    a: NamedArray,
    axis: AxisSelector | Sequence[AxisSelector] | AxisMapping | None = None,
    norm: str | None = None,
) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.ihfft`.

    Only a single axis is supported; passing a dictionary with more than one
    entry will raise an error.
    """

    if isinstance(axis, Mapping):
        if len(axis) != 1:
            raise ValueError("ihfft only supports a single axis")
        key, val = next(iter(axis.items()))
        if isinstance(val, Axis):
            axis = val
        elif val is None:
            axis = key
        else:
            name = key.name if isinstance(key, Axis) else key
            axis = Axis(name, int(val))

    idx, ax, n = _single_axis(a, axis)  # type: ignore[arg-type]
    out = jfft.ihfft(a.array, n=n, axis=idx, norm=norm)
    length = n if n is not None else ax.size // 2 + 1
    new_axis = ax.resize(length)
    axes = list(a.axes)
    axes[idx] = new_axis
    return NamedArray(out, tuple(axes))


def fftshift(x: NamedArray, axes: AxisSelection | None = None) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.fftshift`."""

    if axes is None:
        out = jfft.fftshift(x.array)
    else:
        idxs = x.axis_indices(axes)
        if isinstance(idxs, tuple):
            if any(i is None for i in idxs):
                raise ValueError(f"Axis {axes} not found in {x.axes}")
        elif idxs is None:
            raise ValueError(f"Axis {axes} not found in {x.axes}")
        out = jfft.fftshift(x.array, axes=idxs)
    return NamedArray(out, x.axes)


def ifftshift(x: NamedArray, axes: AxisSelection | None = None) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.ifftshift`."""

    if axes is None:
        out = jfft.ifftshift(x.array)
    else:
        idxs = x.axis_indices(axes)
        if isinstance(idxs, tuple):
            if any(i is None for i in idxs):
                raise ValueError(f"Axis {axes} not found in {x.axes}")
        elif idxs is None:
            raise ValueError(f"Axis {axes} not found in {x.axes}")
        out = jfft.ifftshift(x.array, axes=idxs)
    return NamedArray(out, x.axes)


def fftfreq(axis: Axis, d: float = 1.0) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.fftfreq`."""

    return NamedArray(jfft.fftfreq(axis.size, d), (axis,))


def rfftfreq(axis: Axis, d: float = 1.0) -> NamedArray:
    """Named version of :func:`jax.numpy.fft.rfftfreq`."""

    new_axis = axis.resize(axis.size // 2 + 1)
    return NamedArray(jfft.rfftfreq(axis.size, d), (new_axis,))


__all__ = [
    "fft",
    "ifft",
    "rfft",
    "irfft",
    "hfft",
    "ihfft",
    "fftfreq",
    "rfftfreq",
    "fftshift",
    "ifftshift",
]



# Copyright 2025 The Levanter Authors
#
# SPDX-License-Identifier: Apache-2.0


# Support for FP8
# Much of this is lifted from FLAX
# https://github.com/google/flax/blob/main/flax/linen/fp8_ops.py
import dataclasses
import functools
import warnings
from dataclasses import dataclass
from typing import Protocol, TypeVar

import aqt.jax.v2.config as aqt_config
import equinox as eqx
import jax
import jax.random as jrandom
from aqt.jax.v2.aqt_dot_general import DotGeneral
from jax import numpy as jnp
from jax.tree_util import DictKey, FlattenedIndexKey, GetAttrKey, SequenceKey
from jaxtyping import DTypeLike, PyTree

import haliax.nn as hnn
from haliax.state_dict import StateDict
from haliax.types import PrecisionLike

from ._src.fp8 import dot_general_with_precision, in_qdq, out_qdq
from .axis import Axis
from .hof import vmap

T = TypeVar("T")


class OverwriteWithGradient(eqx.Module):
    """
    Sometimes there is state that must be computed in the backward pass which we want to
    persist for subsequent passes. Typically, we see this with quantization, particularly
    FP8. This module is a marker that indicates to [haliax.quantization.apply_updates][] that the
    gradient should be used to overwrite the state rather than added to it.

    Typically this is used in conjunction with [haliax.quantization.partition_for_grad_overwrite][]
    and the types are kinds of DotGeneralOp.
    """

    pass


def partition_for_grad_overwrite(grad: T) -> tuple[T, T]:
    """
    This function is used to partition the state of a module into two parts: one that will be
    overwritten by the gradient and one that will be updated by the gradient. This is used by
    [equinox.apply_updates][] to determine which state should be updated and which should
    be overwritten.
    The usual pattern is something like:

        ```python
        grads = jax.grad(loss_fn)(model)
        overwrites, grads = partition_for_grad_overwrite(grads)
        updates = optimizer.update(grads, params=model)
        model = hax.quant.apply_updates(model, updates, overwrites)
        ```

    """

    def is_overwrite_with_gradient(v):
        return isinstance(v, OverwriteWithGradient)

    x, y = eqx.partition(grad, is_overwrite_with_gradient, is_leaf=is_overwrite_with_gradient)
    return x, y


def apply_updates(tree, updates, overwrites):
    """
    A `jax.tree_util.tree_map`-broadcasted version of
    ```python
    if overwrite is not None:
        return overwrite
    if update is None:
        return model
    else:
        return model + update
    """

    def _apply_update(tree, update, overwrite):
        if overwrite is not None:
            return overwrite

        return eqx.apply_updates(tree, update)

    def is_leaf(x):
        return x is None or isinstance(x, OverwriteWithGradient)

    return jax.tree_util.tree_map(_apply_update, tree, updates, overwrites, is_leaf=is_leaf)


class DotGeneralOp(Protocol):
    """
    This protocol is used to define the signature of the `dot_general` function that is
    passed to the `Linear` module. This is used to allow for custom dot_general functions
    for quantized types.
    """

    def __call__(
        self,
        lhs,
        rhs,
        dimension_numbers,
        precision: PrecisionLike = None,
        preferred_element_type: DTypeLike | None = None,
    ) -> jnp.ndarray: ...

    @staticmethod
    def default():
        return DefaultDotGeneralOp.init()


class DefaultDotGeneralOp(eqx.Module):
    """
    The default dot_general function that is used by the `Linear` module. This is the
    standard JAX `jax.lax.dot_general` function.

    Notes:
        We could have used `jax.lax.dot_general` directly, but we use this class so that we don't
        unnecessarily have functions as leaves in the module tree.
    """

    def __call__(
        self,
        lhs,
        rhs,
        dimension_numbers,
        precision: PrecisionLike = None,
        preferred_element_type: DTypeLike | None = None,
    ) -> jnp.ndarray:
        return jax.lax.dot_general(lhs, rhs, dimension_numbers, precision, preferred_element_type)

    # not really necessary, but it's nice to have a singleton
    @staticmethod
    def init():
        if not hasattr(DefaultDotGeneralOp, "_instance"):
            DefaultDotGeneralOp._instance = DefaultDotGeneralOp()

        return DefaultDotGeneralOp._instance


class Fp8DotGeneralOp(OverwriteWithGradient):
    input_scale: jnp.ndarray
    output_grad_scale: jnp.ndarray
    kernel_scale: jnp.ndarray
    input_amax_history: jnp.ndarray
    output_grad_amax_history: jnp.ndarray
    kernel_amax_history: jnp.ndarray
    compute_dtype: DTypeLike | None = eqx.field(static=True)

    @classmethod
    def init(cls, amax_history_length: int = 1024, compute_dtype: DTypeLike = None):
        return cls(
            input_scale=jnp.ones(1, dtype=jnp.float32),
            output_grad_scale=jnp.ones(1, dtype=jnp.float32),
            kernel_scale=jnp.ones(1, dtype=jnp.float32),
            input_amax_history=jnp.zeros(amax_history_length, dtype=jnp.float32),
            output_grad_amax_history=jnp.zeros(amax_history_length, dtype=jnp.float32),
            kernel_amax_history=jnp.zeros(amax_history_length, dtype=jnp.float32),
            compute_dtype=compute_dtype,
        )

    # copied from flax
    def __call__(
        self,
        lhs,
        rhs,
        dimension_numbers,
        precision: PrecisionLike = None,
        preferred_element_type: DTypeLike | None = None,
    ):
        # Use the `k.dtype` since it aligns with the `dtype` of its layers,
        # namely, the computation data type.
        if self.compute_dtype is None:
            comp_dtype = rhs.dtype
        else:
            comp_dtype = self.compute_dtype
        lhs = jnp.asarray(lhs, comp_dtype)

        x_qdq = in_qdq(comp_dtype, lhs, self.input_scale, self.input_amax_history)
        k_qdq = in_qdq(comp_dtype, rhs, self.kernel_scale, self.kernel_amax_history)
        y_qdq = dot_general_with_precision(x_qdq, k_qdq, dimension_numbers, precision, preferred_element_type)
        y = out_qdq(comp_dtype, y_qdq, self.output_grad_scale, self.output_grad_amax_history)

        return y


class Int8DotGeneralOp(OverwriteWithGradient):

    cfg: DotGeneral

    @classmethod
    def init(cls):
        cfg = aqt_config.config_v3()
        return cls(cfg)

    def __call__(
        self,
        lhs,
        rhs,
        dimension_numbers,
        precision,
        preferred_element_type=None,
    ):
        cfg = aqt_config.set_context(self.cfg, jrandom.PRNGKey(42), train_step=None)
        return cfg(lhs, rhs, dimension_numbers, precision, preferred_element_type)

    def to_state_dict(tree: PyTree, prefix: str | None = None) -> StateDict:
        warnings.warn("Ignore all int8 states (if any) for now.")
        return {}


@dataclass(frozen=True)
class QuantizationConfig:
    targets: list[str] | str | None = dataclasses.field(default=None)
    """
    If provided, only modules with names in this list will be quantized. If a single string, will be treated as a regex
    """

    amax_history_length: int = 1024
    compute_dtype: DTypeLike = None

    fp8: bool = False
    int8: bool = False

    def __post_init__(self):
        assert not (self.fp8 and self.int8), "Cannot use FP8 and INT8 quantization at the same time."


def quantize_linear_layers(tree: T, config: QuantizationConfig) -> T:
    """
    Converts a module tree to use FP8/INT8 quantization.
    """
    if config.fp8:
        return _quantize_linear_layers(tree, config, Fp8DotGeneralOp, config.amax_history_length, config.compute_dtype)
    elif config.int8:
        return _quantize_linear_layers(tree, config, Int8DotGeneralOp)
    else:
        warnings.warn("Both fp8 and int8 are set to False. `quantize_linear_layers()` is no-op.")
        return tree


def _quantize_linear_layers(tree: T, config: QuantizationConfig, dot_general_cls, *args, **kwargs) -> T:
    """
    Linear modules that have a name that matches the targets (if provided) will be converted to quantized version.
    (If targets is None, all linear modules will be converted.)

    This essentially goes through and adds corresponding DotGeneralOp to the Linear modules.
    """

    def _is_special_module(module):
        # TODO: add conv?
        return isinstance(module, hnn.Linear) or isinstance(module, hnn.Stacked)

    def _batchify_ctor(ctor, batch_dims):
        # this is gross but it basically just vmaps the ctor over each batch dimension
        return functools.reduce(lambda ctor, batch_axis: vmap(ctor, batch_axis), reversed(batch_dims), ctor)

    # TODO: test scanlayers for dg
    def quantize_module(path_prefix, batch_dims: tuple[Axis, ...], path, module: T) -> T:
        path = path_prefix + path
        if isinstance(module, hnn.Stacked):
            new_inner = jax.tree_util.tree_map_with_path(
                functools.partial(quantize_module, path_prefix + (GetAttrKey("stacked"),), batch_dims + (module.Block,)),  # type: ignore
                module.stacked,
                is_leaf=_is_special_module,
            )
            return dataclasses.replace(module, stacked=new_inner)  # type: ignore
        elif isinstance(module, hnn.Linear):
            if _matches_target(path, config):
                vmapped_dg = _batchify_ctor(dot_general_cls.init, batch_dims)(*args, **kwargs)
                module = dataclasses.replace(module, dot_general=vmapped_dg)  # type: ignore
            return module
        else:
            return module

    return jax.tree_util.tree_map_with_path(
        lambda p, m: quantize_module((), (), p, m), tree, is_leaf=_is_special_module
    )


def _matches_target(key_path, config: QuantizationConfig) -> bool:
    if not key_path:
        key = ""
    else:
        key = _key_path_to_str(key_path[-1:])

    if config.targets is None:
        return True
    if isinstance(config.targets, list):
        return key in config.targets

    import re

    key_path_str = _key_path_to_str(key_path)
    return re.match(config.targets, key_path_str) is not None


def _key_path_to_str(key_path: tuple) -> str:
    out = ""
    for k in key_path:
        match k:
            case SequenceKey(i):  # type: ignore
                out = _join_key(out, str(i))
            case GetAttrKey(name):  # type: ignore
                out = _join_key(out, name)
            case DictKey(key):  # type: ignore
                out = _join_key(out, key)
            case FlattenedIndexKey(i):  # type: ignore
                out = _join_key(out, str(i))
            case _:
                warnings.warn(f"Unsupported key type {k}")
                out = _join_key(out, str(k))
    return out


def _join_key(prefix: str, key: str) -> str:
    if prefix:
        return f"{prefix}.{key}"
    return key
